/*
 * @(#)Welltris.c
 *
 * Copyright 1993 - 2010  David A. Bagley, bagleyd@tux.org
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995  Q. Alex Zhao, azhao@cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Methods file for Welltris */

#include "file.h"
#include "rngs.h"
#include "sound.h"
#include "WelltrisP.h"

/*
static Point big_trapazoid[MAX_SIDES][MAX_SIDES] =
	{{4 * DELTA, 2 * DELTA},
	{w->welltris.endPt - 8 * DELTA, 0},
	{4 * DELTA - w->welltris.baseWidth, w->welltris.baseWidth - 4 * DELTA},
	{-w->welltris.baseWidth, 0},
	{w->welltris.endPt - 2 * DELTA, 3 * DELTA},
	{0, w->welltris.endPt - 7 * DELTA},
	{4 * DELTA - w->welltris.baseWidth, 4 * DELTA - w->welltris.baseWidth},
	{0, -w->welltris.baseWidth - DELTA},
	{w->welltris.endPt - 3 * DELTA, w->welltris.endPt - 2 * DELTA},
	{6 * DELTA - w->welltris.endPt, 0},
	{w->welltris.baseWidth - 4 * DELTA, 4 * DELTA - w->welltris.baseWidth},
	{w->welltris.baseWidth + 2 * DELTA, 0},
	{2 * DELTA, w->welltris.endPt - 4 * DELTA},
	{0, 7 * DELTA - w->welltris.endPt},
	{w->welltris.baseWidth - 4 * DELTA, w->welltris.baseWidth - 4 * DELTA},
	{0, w->welltris.baseWidth + DELTA}};
*/

#ifdef WINVER
#ifndef LOGPATH
#define LOGPATH "/usr/tmp"
#endif
#ifndef INIFILE
#define INIFILE "wwelltris.ini"
#endif

#define SECTION "setup"
#else

#if defined( USE_SOUND ) && defined( USE_NAS )
Display *dsp;
#endif

#ifndef LOGPATH
#ifdef VMS
#define LOGPATH "SYS$SCRATCH:"
#else
#define LOGPATH "/usr/tmp"
#endif
#endif

static unsigned char gray1_bits[] = {0x01, 0x02};
static Pixmap graypix;

static void resizeGame(WelltrisWidget w);
static void sizeGame(WelltrisWidget w);
static Boolean setValuesGame(Widget current, Widget request, Widget renew);
static void destroyGame(Widget old);
static void quitGame(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void initializeGame(Widget request, Widget renew);
static void exposeGame(Widget renew, XEvent *event, Region region);
static void mouseGame(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void hideGame(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void clearGame(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void enterGame(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void leaveGame(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void reflectGame(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void rotateGame(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void moveGameLeft(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void moveGameRight(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void dropGame(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void moveGameInput(WelltrisWidget w, int direction);
static void incrementGameLevel(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void decrementGameLevel(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void incrementGameUnits(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void decrementGameUnits(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void incrementGameDiameter(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void decrementGameDiameter(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void fixedfillGame(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void suddenAppearGame(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void diagonalGame(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void mixedGame(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void bonusGame(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void cwGame(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void plainPieceGame(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void showNextGame(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);
static void gridGame(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void soundGame(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void pauseGame(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void fill0Game(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void fill1Game(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void fill2Game(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void fill3Game(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void fill4Game(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void fill5Game(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void fill6Game(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void fill7Game(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void fill8Game(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void fill9Game(WelltrisWidget w, XEvent *event, char **args, int nArgs);
static void fill10Game(WelltrisWidget w,
	XEvent *event, char **args, int nArgs);

static char defaultTranslationsGame[] =
"<KeyPress>q: Quit()\n\
 Ctrl<KeyPress>C: Quit()\n\
 <KeyPress>osfCancel: Hide()\n\
 <KeyPress>Escape: Hide()\n\
 <KeyPress>osfEscape: Hide()\n\
 Ctrl<KeyPress>[: Hide()\n\
 <KeyPress>0x1B: Hide()\n\
 <Btn4Down>: Mouse() Reflect()\n\
 <KeyPress>i: Reflect()\n\
 <KeyPress>e: Reflect()\n\
 <KeyPress>F11: Reflect()\n\
 <KeyPress>KP_Divide: Reflect()\n\
 <KeyPress>R5: Reflect()\n\
 <Btn2Down>: Mouse() Rotate()\n\
 <KeyPress>k: Rotate()\n\
 <KeyPress>d: Rotate()\n\
 <KeyPress>Up: Rotate()\n\
 <KeyPress>osfUp: Rotate()\n\
 <KeyPress>KP_Up: Rotate()\n\
 <KeyPress>R8: Rotate()\n\
 <KeyPress>KP_8: Rotate()\n\
 <KeyPress>F12: Rotate()\n\
 <KeyPress>Begin: Rotate()\n\
 <KeyPress>KP_5: Rotate()\n\
 <KeyPress>R11: Rotate()\n\
 <Btn1Down>: Mouse() MoveLeft()\n\
 <KeyPress>j: MoveLeft()\n\
 <KeyPress>s: MoveLeft()\n\
 <KeyPress>Left: MoveLeft()\n\
 <KeyPress>osfLeft: MoveLeft()\n\
 <KeyPress>KP_Left: MoveLeft()\n\
 <KeyPress>KP_4: MoveLeft()\n\
 <KeyPress>R10: MoveLeft()\n\
 <Btn3Down>: Mouse() MoveRight()\n\
 <KeyPress>l: MoveRight()\n\
 <KeyPress>f: MoveRight()\n\
 <KeyPress>Right: MoveRight()\n\
 <KeyPress>osfRight: MoveRight()\n\
 <KeyPress>KP_Right: MoveRight()\n\
 <KeyPress>KP_6: MoveRight()\n\
 <KeyPress>R12: MoveRight()\n\
 <Btn5Down>: Mouse() Drop()\n\
 <KeyPress>Return: Drop()\n\
 <KeyPress>KP_Enter: Drop()\n\
 <KeyPress>Down: Drop()\n\
 <KeyPress>osfDown: Drop()\n\
 <KeyPress>KP_Down: Drop()\n\
 <KeyPress>KP_2: Drop()\n\
 <KeyPress>R14: Drop()\n\
 <KeyPress>KP_Space: Drop()\n\
 <KeyPress>space: Drop()\n\
 <Btn4Down>: Mouse() Rotate()\n\
 <Btn5Down>: Mouse() Drop()\n\
 <KeyPress>c: Clear()\n\
 <KeyPress>>: IncrementLevel()\n\
 <KeyPress>.: IncrementLevel()\n\
 <KeyPress><: DecrementLevel()\n\
 <KeyPress>+: IncrementUnits()\n\
 <KeyPress>=: IncrementUnits()\n\
 <KeyPress>-: DecrementUnits()\n\
 <KeyPress>_: DecrementUnits()\n\
 <KeyPress>]: IncrementDiameter()\n\
 <KeyPress>}: IncrementDiameter()\n\
 <KeyPress>[: DecrementDiameter()\n\
 <KeyPress>{: DecrementDiameter()\n\
 <KeyPress>x: FixedFill()\n\
 <KeyPress>r: SuddenAppear()\n\
 <KeyPress>g: Diagonal()\n\
 <KeyPress>m: Mixed()\n\
 <KeyPress>b: Bonus()\n\
 <KeyPress>w: Cw()\n\
 <KeyPress>p: PlainPiece()\n\
 <KeyPress>n: ShowNext()\n\
 Shift<KeyPress>3: Grid()\n\
 Shift<KeyPress>2: Sound()\n\
 <KeyPress>|: Pause()\n\
 <KeyPress>`: Fill0()\n\
 <KeyPress>1: Fill1()\n\
 <KeyPress>2: Fill2()\n\
 <KeyPress>3: Fill3()\n\
 <KeyPress>4: Fill4()\n\
 <KeyPress>5: Fill5()\n\
 <KeyPress>6: Fill6()\n\
 <KeyPress>7: Fill7()\n\
 <KeyPress>8: Fill8()\n\
 <KeyPress>9: Fill9()\n\
 <KeyPress>0: Fill10()\n\
 <EnterWindow>: Enter()\n\
 <LeaveWindow>: Leave()";

/*
 <KeyPress>,: DecrementLevel()\n\
#if defined(UP_REFLECT)
 <KeyPress>KP_Up: Reflect()\n\
 <KeyPress>R8: Reflect()\n\
 <KeyPress>KP_8: Reflect()\n\
#endif
*/

static XtActionsRec actionsListGame[] =
{
	{(char *) "Quit", (XtActionProc) quitGame},
	{(char *) "Hide", (XtActionProc) hideGame},
	{(char *) "Mouse", (XtActionProc) mouseGame},
	{(char *) "Reflect", (XtActionProc) reflectGame},
	{(char *) "Rotate", (XtActionProc) rotateGame},
	{(char *) "MoveLeft", (XtActionProc) moveGameLeft},
	{(char *) "MoveRight", (XtActionProc) moveGameRight},
	{(char *) "Drop", (XtActionProc) dropGame},
	{(char *) "Clear", (XtActionProc) clearGame},
	{(char *) "IncrementLevel", (XtActionProc) incrementGameLevel},
	{(char *) "DecrementLevel", (XtActionProc) decrementGameLevel},
	{(char *) "IncrementUnits", (XtActionProc) incrementGameUnits},
	{(char *) "DecrementUnits", (XtActionProc) decrementGameUnits},
	{(char *) "IncrementDiameter", (XtActionProc) incrementGameDiameter},
	{(char *) "DecrementDiameter", (XtActionProc) decrementGameDiameter},
	{(char *) "FixedFill", (XtActionProc) fixedfillGame},
	{(char *) "SuddenAppear", (XtActionProc) suddenAppearGame},
	{(char *) "Diagonal", (XtActionProc) diagonalGame},
	{(char *) "Mixed", (XtActionProc) mixedGame},
	{(char *) "Bonus", (XtActionProc) bonusGame},
	{(char *) "Cw", (XtActionProc) cwGame},
	{(char *) "PlainPiece", (XtActionProc) plainPieceGame},
	{(char *) "ShowNext", (XtActionProc) showNextGame},
	{(char *) "Grid", (XtActionProc) gridGame},
	{(char *) "Sound", (XtActionProc) soundGame},
	{(char *) "Pause", (XtActionProc) pauseGame},
	{(char *) "Fill0", (XtActionProc) fill0Game},
	{(char *) "Fill1", (XtActionProc) fill1Game},
	{(char *) "Fill2", (XtActionProc) fill2Game},
	{(char *) "Fill3", (XtActionProc) fill3Game},
	{(char *) "Fill4", (XtActionProc) fill4Game},
	{(char *) "Fill5", (XtActionProc) fill5Game},
	{(char *) "Fill6", (XtActionProc) fill6Game},
	{(char *) "Fill7", (XtActionProc) fill7Game},
	{(char *) "Fill8", (XtActionProc) fill8Game},
	{(char *) "Fill9", (XtActionProc) fill9Game},
	{(char *) "Fill10", (XtActionProc) fill10Game},
	{(char *) "Enter", (XtActionProc) enterGame},
	{(char *) "Leave", (XtActionProc) leaveGame}
};

static XtResource resourcesGame[] =
{
	/* 20 * 8 * 3 + 6 */
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(WelltrisWidget, core.width),
	 XtRString, (caddr_t) "486"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(WelltrisWidget, core.height),
	 XtRString, (caddr_t) "486"},
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverseVideo, XtCReverseVideo, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(WelltrisWidget, welltris.foreground),
	 XtRString, (caddr_t) XtDefaultForeground},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(WelltrisWidget, welltris.background),
	 XtRString, (caddr_t) "#AEB2C3" /*XtDefaultBackground*/},
	{XtNdelay, XtCDelay, XtRInt, sizeof (int),
	 XtOffset(WelltrisWidget, welltris.delay),
	 XtRString, (caddr_t) "1000"},
	{XtNsound, XtCSound, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.sound),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbumpSound, XtCBumpSound, XtRString, sizeof (String),
	 XtOffset(WelltrisWidget, welltris.bumpSound),
	 XtRString, (caddr_t) BUMPSOUND},
	{XtNdripSound, XtCDripSound, XtRString, sizeof (String),
	 XtOffset(WelltrisWidget, welltris.dripSound),
	 XtRString, (caddr_t) DRIPSOUND},
	{XtNdiameter, XtCDiameter, XtRInt, sizeof (int),
	 XtOffset(WelltrisWidget, welltris.diameter),
	 XtRString, (caddr_t) "8"},
	{XtNlevel, XtCLevel, XtRInt, sizeof (int),
	 XtOffset(WelltrisWidget, welltris.level),
	 XtRString, (caddr_t) "3"},
	{XtNunits, XtCUnits, XtRInt, sizeof (int),
	 XtOffset(WelltrisWidget, welltris.units),
	 XtRString, (caddr_t) "4"},
	{XtNfill, XtCFill, XtRInt, sizeof (int),
	 XtOffset(WelltrisWidget, welltris.fill),
	 XtRString, (caddr_t) "0"},
	{XtNfixedFill, XtCFixedFill, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.fixedFill),
	 XtRString, (caddr_t) "TRUE"},
	{XtNsuddenAppear, XtCSuddenAppear, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.suddenAppear),
	 XtRString, (caddr_t) "FALSE"},
	{XtNdiagonal, XtCDiagonal, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.diagonal),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmixed, XtCMixed, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.mixed),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbonus, XtCBonus, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.bonus),
	 XtRString, (caddr_t) "TRUE"},
	{XtNcw, XtCCw, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.cw),
	 XtRString, (caddr_t) "TRUE"},
	{XtNplainPiece, XtCPlainPiece, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.plainPiece),
	 XtRString, (caddr_t) "FALSE"},
	{XtNshowNext, XtCShowNext, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.showNext),
	 XtRString, (caddr_t) "TRUE"},
	{XtNgrid, XtCGrid, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.grid),
	 XtRString, (caddr_t) "TRUE"},
	{XtNscore, XtCBoolean, XtRInt, sizeof (int),
	 XtOffset(WelltrisWidget, welltris.score),
	 XtRString, (caddr_t) "0"},
	{XtNcompletedRows, XtCBoolean, XtRInt, sizeof (int),
	 XtOffset(WelltrisWidget, welltris.completedRows),
	 XtRString, (caddr_t) "0"},
	{XtNuserName, XtCUserName, XtRString, sizeof (String),
	 XtOffset(WelltrisWidget, welltris.userName),
	 XtRString, (caddr_t) ""},
	{XtNscoreFile, XtCScoreFile, XtRString, sizeof (String),
	 XtOffset(WelltrisWidget, welltris.scoreFile),
	 XtRString, (caddr_t) ""},
	{XtNscoreOnly, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.scoreOnly),
	 XtRString, (caddr_t) "FALSE"},
	{XtNversionOnly, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.versionOnly),
	 XtRString, (caddr_t) "FALSE"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNact, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(WelltrisWidget, welltris.act),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(WelltrisWidget, welltris.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNpixmapSize, XtCPixmapSize, XtRInt, sizeof (int),
	 XtOffset(WelltrisWidget, welltris.pixmapSize),
	 XtRString, (caddr_t) "64"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(WelltrisWidget, welltris.select),
	 XtRCallback, (caddr_t) NULL}
};

WelltrisClassRec welltrisClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "Welltris",	/* class name */
		sizeof (WelltrisRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) initializeGame,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListGame,	/* actions */
		XtNumber(actionsListGame),	/* num actions */
		resourcesGame,	/* resources */
		XtNumber(resourcesGame),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) destroyGame,	/* destroy */
		(XtWidgetProc) resizeGame,	/* resize */
		(XtExposeProc) exposeGame,	/* expose */
		(XtSetValuesFunc) setValuesGame,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsGame,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass welltrisWidgetClass = (WidgetClass) & welltrisClassRec;

static void
setGame(WelltrisWidget w, int reason)
{
	welltrisCallbackStruct cb;

	cb.reason = reason;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}

static void
setGameNext(WelltrisWidget w, int reason,
		int randomNumber, int randomRotation, int randomReflection,
		int units, Boolean diagonal, Boolean mixed, Boolean newBonus)
{
	welltrisCallbackStruct cb;

	cb.reason = reason;
	cb.randomNumber = randomNumber;
	cb.randomRotation = randomRotation;
	cb.randomReflection = randomReflection;
	cb.units = units;
	cb.diagonal = diagonal;
	cb.mixed = mixed;
	cb.newBonus = newBonus;
	XtCallCallbacks((Widget) w, (char *) XtNselectCallback, &cb);
}

#endif

#include "image.h"

static Pixmap dr = 0; /* dummy for future double buffering */

static const int delay[MAX_LEVELS] = {
	100, 92, 84, 75, 68, 60, 52, 45, 39, 32,
	26, 21, 16, 12, 8, 5, 3, 2, 1, 0
};

static void
checkThings(WelltrisWidget w)
{
	char *buf1 = NULL, *buf2 = NULL;

	if (w->welltris.level < MIN_LEVELS) {
		w->welltris.level = MIN_LEVELS;
		intCat(&buf1, "Number of levels out of bounds, use at least ",
			MIN_LEVELS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->welltris.level);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	} else if (w->welltris.level > MAX_LEVELS) {
		w->welltris.level = MAX_LEVELS;
		intCat(&buf1, "Number of levels out of bounds, use at most ",
			MAX_LEVELS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->welltris.level);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	if (w->welltris.units < MIN_UNITS) {
		w->welltris.units = MIN_UNITS;
		w->welltris.next.units = w->welltris.units;
		intCat(&buf1, "Number of units out of bounds, use at least ",
			MIN_UNITS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->welltris.units);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	} else if (w->welltris.units > MAX_UNITS) {
		w->welltris.units = MAX_UNITS;
		w->welltris.next.units = w->welltris.units;
		intCat(&buf1, "Number of units out of bounds, use at most ",
			MAX_UNITS);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->welltris.units);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	if (w->welltris.diameter < MIN_DIAMETER) {
		w->welltris.diameter = DEFAULT_DIAMETER;
		intCat(&buf1, "Diameter out of bounds, use at least ",
			MIN_DIAMETER);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->welltris.diameter);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
	if (!LEGAL(w->welltris.units, w->welltris.diagonal,
			w->welltris.mixed, FALSE)) {
		intCat(&buf1, "Illegal configuration of units ",
			w->welltris.units);
		stringCat(&buf2, buf1, ", diagonal ");
		free(buf1);
		intCat(&buf1, buf2, (int) w->welltris.diagonal);
		free(buf2);
		stringCat(&buf2, buf1, ", mixed ");
		free(buf1);
		intCat(&buf1, buf2, (int) w->welltris.mixed);
		free(buf2);
		stringCat(&buf2, buf1, ", defaulting to units ");
		free(buf1);
		intCat(&buf1, buf2, DEFAULT_UNITS);
		free(buf2);
		stringCat(&buf2, buf1, ", diagonal ");
		free(buf1);
		intCat(&buf1, buf2, (int) DEFAULT_DIAGONAL);
		free(buf2);
		stringCat(&buf2, buf1, ", mixed ");
		free(buf1);
		intCat(&buf1, buf2, (int) DEFAULT_MIXED);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
		w->welltris.units = DEFAULT_UNITS;
		w->welltris.diagonal = FALSE;
		w->welltris.mixed = FALSE;
		w->welltris.next.units = w->welltris.units;
		w->welltris.next.diagonal = w->welltris.diagonal;
		w->welltris.next.mixed = w->welltris.mixed;
	}
	if (w->welltris.fill < 0) {
		w->welltris.fill = 0;
		DISPLAY_WARNING("Number for fill can not be negative, defaulting to 0");
	} else if (w->welltris.fill > MAX_FILL) {
		w->welltris.fill = MAX_FILL;
		intCat(&buf1, "Number for fill out of bounds, use at most ",
			MAX_FILL);
		stringCat(&buf2, buf1, ", defaulting to ");
		free(buf1);
		intCat(&buf1, buf2, w->welltris.fill);
		free(buf2);
		DISPLAY_WARNING(buf1);
		free(buf1);
	}
}

static void
wallToBase(WelltrisWidget w, int *base_x, int *base_y,
		int wall_x, int wall_y)
{
	switch (wall_x / w->welltris.diameter) {
	case 0:
		*base_x = wall_x;
		*base_y = wall_y - w->welltris.depth;
		break;
	case 1:
		*base_x = w->welltris.diameter - 1 + w->welltris.depth - wall_y;
		*base_y = wall_x - w->welltris.diameter;
		break;
	case 2:
		*base_x = w->welltris.perimeter - 1 - w->welltris.diameter - wall_x;
		*base_y = w->welltris.diameter - 1 + w->welltris.depth - wall_y;
		break;
	case 3:
		*base_x = wall_y - w->welltris.depth;
		*base_y = w->welltris.perimeter - 1 - wall_x;
		break;
	default:
		(void) fprintf (stderr,
			"wallToBase kinds range 0-3, got %d.\n",
			wall_x / w->welltris.diameter);
		exit(-1);
	}
}

static int
checkWall(WelltrisWidget w, int pmid, int base_i, int base_j)
{
	int newPmid = pmid;

	if (newPmid < 0)
		return newPmid;
	/* Check if next to well wall */
	if (base_i == 0)
		newPmid = CHECKLEFT((unsigned int) newPmid);
	if (base_j == 0)
		newPmid = CHECKUP((unsigned int) newPmid);
	if (base_i == w->welltris.diameter - 1)
		newPmid = CHECKRIGHT((unsigned int) newPmid);
	if (base_j == w->welltris.diameter - 1)
		newPmid = CHECKDOWN((unsigned int) newPmid);
	return newPmid;
}

/* static int
checkSides(WelltrisWidget w, int pmid, int i)
{
	int newPmid = pmid;

	if (newPmid < 0)
		return newPmid;
	if (i == 0 || i == 3 * w->welltris.diameter - 1)
		newPmid = CHECKLEFT((unsigned int) newPmid);
	else if (i == 2 * w->welltris.diameter || i == w->welltris.diameter - 1)
		newPmid = CHECKRIGHT((unsigned int) newPmid);
	else if (i == w->welltris.diameter || i == 4 * w->welltris.diameter - 1)
		newPmid = CHECKUP((unsigned int) newPmid);
	else if (i == 3 * w->welltris.diameter || i == 2 * w->welltris.diameter - 1)
		newPmid = CHECKDOWN((unsigned int) newPmid);
	return newPmid;
}

static int
checkBottom(WelltrisWidget w, int pmid, int i, int j)
{
	if (pmid < 0 || j != w->welltris.depth - 1)
		return pmid;
	switch (i / w->welltris.diameter) {
	case 0:
		return CHECKDOWN((unsigned int) pmid);
	case 1:
		return CHECKLEFT((unsigned int) pmid);
	case 2:
		return CHECKUP((unsigned int) pmid);
	case 3:
		return CHECKRIGHT((unsigned int) pmid);
	}
	return EMPTY;
}*/

static int
checkWallSides(WelltrisWidget w, int pmid, int i)
{
	int newPmid = pmid;

	if (newPmid < 0)
		return newPmid;
	if (i % w->welltris.diameter == 0)
		newPmid = CHECKLEFT((unsigned int) newPmid);
	else if (i % w->welltris.diameter == w->welltris.diameter - 1)
		newPmid = CHECKRIGHT((unsigned int) newPmid);
	return newPmid;
}

static int
checkWallBottom(WelltrisWidget w, int pmid, int j)
{
	if (pmid < 0 || j != w->welltris.depth - 1)
		return pmid;
	return CHECKDOWN((unsigned int) pmid);
}

static int
rotateIndex(WelltrisWidget w, int pmid, int i, int j)
{
	int wallRotation, pm1, pm0;

	if (pmid == EMPTY)
		return EMPTY;
	if (j < w->welltris.depth)
		return pmid;
	wallRotation = (MAX_SIDES - i / w->welltris.diameter) % MAX_SIDES;
	pm1 = (0xF0 & pmid) << wallRotation;
	pm0 = (0xF & pmid) << wallRotation;
	pm1 = (0xF0 & pm1) | ((pm1 & 0xF00) >> MAX_SIDES);
	pm0 = (0xF & pm0) | ((pm0 & 0xF0) >> MAX_SIDES);
	return (pm1 | pm0);
}

#ifdef BOTTOM_BORDER
static void
drawUnitBorder(WelltrisWidget w, int i, int j)
{
	int baseWidth = w->welltris.diameter * w->welltris.boxSize;
	GC gc;

	if (w->welltris.grid)
		gc = w->welltris.inverseGC[3];
	else
		gc = w->welltris.inverseGC[2];
	DRAWRECTANGLE(w, dr, gc,
		w->welltris.boxSize * i + baseWidth + w->welltris.offset.x + 1,
		w->welltris.boxSize * j + baseWidth + w->welltris.offset.y + 1,
		w->welltris.boxSize - 1, w->welltris.boxSize - 1);
}
#endif

static void
drawUnit(WelltrisWidget w, int pmid, int cid, int i, int j)
{
#ifdef IMAGE
	Boolean image = True;
#ifdef WINVER
	HBITMAP hOldBitmap, monoOldBitmap, hBitmap;
	HDC monoDC;
#endif
#endif
	int baseWidth = w->welltris.diameter * w->welltris.boxSize;

	if (i < 0 || i >= w->welltris.diameter ||
			j < 0 || j >= w->welltris.diameter)
		return;
#ifdef IMAGE
	if (images[FALSE][pmid % BITMAPS] == None) {
		if (images[FALSE][BITMAPS - 1] != None)
			(void) printf("Bitmap unknown %d.\n", pmid);
		image = False;
	}
#ifdef WINVER
	if (image && (monoDC = CreateCompatibleDC(0)) == NULL) {
		(void) printf("Can not create mono DC.\n");
		image = False;
	}
	if (image && (w->core.memDC =
			CreateCompatibleDC(w->core.hDC)) == NULL) {
		(void) printf("Can not create compatible DC.\n");
		image = False;
	}
	if (image && (hBitmap = CreateCompatibleBitmap(w->core.hDC,
			w->welltris.boxSize, w->welltris.boxSize)) == NULL) {
		(void) printf("Can not create compatible bitmap.\n");
		image = False;
	}
#endif
	if (image) {
#ifdef WINVER
		SetTextColor(w->core.memDC, w->welltris.inverseGC[2]);
		SetBkColor(w->core.memDC, w->welltris.thingGC[cid][2]);
		FILLRECTANGLE(w, hBitmap, w->welltris.thingGC[cid][2],
			0, 0,
			w->welltris.boxSize, w->welltris.boxSize);
		monoOldBitmap = (HBITMAP) SelectObject(monoDC,
			images[FALSE][pmid % BITMAPS]);
		hOldBitmap = SelectObject(w->core.memDC, hBitmap);
		BitBlt(w->core.memDC,
			0, 0, w->welltris.boxSize, w->welltris.boxSize,
			monoDC, 0, 0,
			SRCCOPY);
		(void) SelectObject(monoDC, monoOldBitmap);
		DeleteDC(monoDC);
		BitBlt(w->core.hDC,
			w->welltris.boxSize * i + baseWidth + w->welltris.offset.x + 1,
			w->welltris.boxSize * j + baseWidth + w->welltris.offset.y + 1,
			w->welltris.boxSize, w->welltris.boxSize,
			w->core.memDC, 0, 0,
			SRCCOPY);
		(void) SelectObject(w->core.memDC, hOldBitmap);
		DeleteObject(hBitmap);
		DeleteDC(w->core.memDC);
#else
		(void) XPutImage(XtDisplay(w), XtWindow(w),
			w->welltris.thingGC[cid][2],
			images[FALSE][pmid % BITMAPS],
			0, 0,
			w->welltris.boxSize * i + baseWidth + w->welltris.offset.x + 1,
			w->welltris.boxSize * j + baseWidth + w->welltris.offset.y + 1,
			w->welltris.boxSize, w->welltris.boxSize);
#endif
	} else
#endif
	{
		FILLRECTANGLE(w, dr, w->welltris.thingGC[cid][2],
			w->welltris.boxSize * i + baseWidth + w->welltris.offset.x + 1,
			w->welltris.boxSize * j + baseWidth + w->welltris.offset.y + 1,
			w->welltris.boxSize, w->welltris.boxSize);
	}
}

static void
clearUnit(WelltrisWidget w, int i, int j)
{
	if (i >= 0 && i < w->welltris.diameter &&
			j >= 0 && j < w->welltris.diameter) {
		int baseWidth = w->welltris.diameter * w->welltris.boxSize;

		FILLRECTANGLE(w, dr, w->welltris.inverseGC[2],
			w->welltris.boxSize * i + baseWidth + w->welltris.offset.x + 1,
			w->welltris.boxSize * j + baseWidth + w->welltris.offset.y + 1,
			w->welltris.boxSize, w->welltris.boxSize);
	}
}

static void
xorUnit(WelltrisWidget w, int i, int j)
{
	int baseWidth = w->welltris.diameter * w->welltris.boxSize;

#ifdef WINVER
	HDC memDC;
	HBITMAP hBitmap, hOldBitmap;

	if ((memDC = CreateCompatibleDC(w->core.hDC)) == NULL) {
		DISPLAY_ERROR("Not enough memory (xor memDC), exiting.");
	}
	if ((hBitmap = CreateCompatibleBitmap(w->core.hDC,
			w->welltris.boxSize,
			w->welltris.boxSize)) == NULL) {
		DISPLAY_ERROR("Not enough memory (xor hBitmap), exiting.");
	}
	hOldBitmap = SelectObject(memDC, hBitmap);
	BitBlt(memDC, 0, 0,
		w->welltris.boxSize, w->welltris.boxSize,
		w->core.hDC,
		w->welltris.boxSize * i + baseWidth + w->welltris.offset.x + 1,
		w->welltris.boxSize * j + baseWidth + w->welltris.offset.y + 1,
		SRCCOPY);
	BitBlt(w->core.hDC,
		w->welltris.boxSize * i + baseWidth + w->welltris.offset.x + 1,
		w->welltris.boxSize * j + baseWidth + w->welltris.offset.y + 1,
		w->welltris.boxSize, w->welltris.boxSize,
		memDC,
		0, 0,
		DSTINVERT);
	(void) SelectObject(memDC, hOldBitmap);
	DeleteObject(hBitmap);
	DeleteDC(memDC);
#else
	FILLRECTANGLE(w, dr, w->welltris.xorGC,
		w->welltris.boxSize * i + baseWidth + w->welltris.offset.x + 1,
		w->welltris.boxSize * j + baseWidth + w->welltris.offset.y + 1,
		w->welltris.boxSize, w->welltris.boxSize);
#endif
}

#if 0
static void
moveUnit(WelltrisWidget w, int ai, int aj, int bi, int bj)
{
#ifdef WINVER
	BitBlt(w->core.hDC,
		w->welltris.offset.x + bi * w->welltris.boxSize + 1,
		w->welltris.offset.y + bj * w->welltris.boxSize + 1,
		w->welltris.boxSize, w->welltris.boxSize,
		w->core.hDC,
		w->welltris.offset.x + ai * w->welltris.boxSize + 1,
		w->welltris.offset.y + aj * w->welltris.boxSize + 1,
		SRCCOPY);
#else
	VOID XSetGraphicsExposures(XtDisplay(w), w->welltris.inverseGC[2],
		False);
	VOID XCopyArea(XtDisplay(w), XtWindow(w), XtWindow(w),
		w->welltris.inverseGC[2],
		w->welltris.offset.x + ai * w->welltris.boxSize + 1,
		w->welltris.offset.y + aj * w->welltris.boxSize + 1,
		w->welltris.boxSize, w->welltris.boxSize,
		w->welltris.offset.x + bi * w->welltris.boxSize + 1,
		w->welltris.offset.y + bj * w->welltris.boxSize + 1);
#endif
}
#endif

static void
fillTrapazoid(WelltrisWidget w,
		int pm, int color,
		int ulx, int uly, int base_x, int base_y,
		int lrx, int lry, int plateau_x, int plateau_y,
		Boolean graphicalXor)
{
	Point trapazoidList[MAX_SIDES + 1];
	GC gc[FG_SHADES];
	int i;

	for (i = 0; i < FG_SHADES; i++)
		gc[i] = w->welltris.thingGC[color][i];
	if (graphicalXor) {
#ifdef WINVER
		if (color == CLEAR_COLOR)
			gc[2] = w->welltris.inverseGC[2];
		else
			gc[2] = w->welltris.thingGC[color][2];
		for (i = 0; i < FG_SHADES; i++)
			gc[i] = RGB(255 - GetRValue(gc[i]),
				255 - GetGValue(gc[i]),
				255 - GetBValue(gc[i]));
#else
		gc[2] = w->welltris.xorGC;
#endif
	} else if (color == CLEAR_COLOR) {
		gc[2] = w->welltris.inverseGC[2];
	} else {
		gc[2] = w->welltris.thingGC[color][2];
#ifndef WINVER
#ifdef IMAGE
		VOID XSetFillStyle(XtDisplay(w), gc[2], FillOpaqueStippled);
		VOID XSetStipple(XtDisplay(w), gc[2], graypix);
#endif
#endif
	}
	trapazoidList[0].x = ulx + w->welltris.offset.x;
	trapazoidList[0].y = uly + w->welltris.offset.y;
	trapazoidList[1].x = base_x;
	trapazoidList[1].y = base_y;
	trapazoidList[2].x = lrx - base_x - ulx;
	trapazoidList[2].y = lry - base_y - uly;
	trapazoidList[3].x = -plateau_x;
	trapazoidList[3].y = -plateau_y;
	trapazoidList[4].x = ulx + plateau_x - lrx;
	trapazoidList[4].y = uly + plateau_y - lry;
	POLYGON((WelltrisWidget) w, dr,
		gc[2], gc[2], trapazoidList, MAX_SIDES, True, False);
#ifndef WINVER
#ifdef IMAGE
	if (!graphicalXor && color != CLEAR_COLOR) {
		VOID XSetFillStyle(XtDisplay(w), gc[2], FillSolid);
	}
#endif
#endif
	/* Code added to use pmid info */
	if (!w->welltris.plainPiece && gc[2] != w->welltris.inverseGC[2] &&
			gc[2] != w->welltris.xorGC) {
		Point points[MAX_SIDES + 1];
		Point midPoint;
		Point edgePoints[MAX_SIDES + 1];
		Point pts[5];
		GC *shade[MAX_SIDES];

		if (pm > BITMAPS) {
			shade[2]= &(gc[1]);
			shade[1]= &(gc[0]);
			shade[0]= &(gc[3]);
			shade[3]= &(gc[4]);
		} else {
			shade[0]= &(gc[1]);
			shade[1]= &(gc[0]);
			shade[2]= &(gc[3]);
			shade[3]= &(gc[4]);
		}
		points[0] = trapazoidList[0];
		midPoint.x = 0;
		midPoint.y = 0;
		for (i = 1; i <= MAX_SIDES; i++) {
			points[i].x = points[i - 1].x + trapazoidList[i].x;
			points[i].y = points[i - 1].y + trapazoidList[i].y;
			midPoint.x += points[i - 1].x;
			midPoint.y += points[i - 1].y;
		}
		midPoint.x /= MAX_SIDES;
		midPoint.y /= MAX_SIDES;
		for (i = 0; i < MAX_SIDES; i++) {
			edgePoints[i].x = (points[i + 1].x + points[i].x) / 2;
			edgePoints[i].y = (points[i + 1].y + points[i].y) / 2;
		}
		for (i = 0; i < MAX_SIDES; i++) {
			int invi = (MAX_SIDES - i) % MAX_SIDES;
			int invi1 = (MAX_SIDES - i + 3) % MAX_SIDES;
			int invi3 = (MAX_SIDES - i + 1) % MAX_SIDES;
			int i1 = (i + 1) % MAX_SIDES;
			int i3 = (i + 3) % MAX_SIDES;

			if ((pm & (1 << invi)) != 0) {
				pts[0] = points[i3];
				pts[1] = points[i];
				pts[2] = midPoint;
				pts[3] = points[i3];
				POLYGON(w, dr, *(shade[i]),
					*(shade[i]), pts, 3, True, True);
			} else {
				if ((pm & (1 << invi3)) != 0 ||
						(pm & (16 << invi)) != 0) {
					pts[0] = points[i3];
					pts[1] = midPoint;
					pts[2] = edgePoints[i3];
					pts[3] = points[i3];
					POLYGON(w, dr, *(shade[i3]),
						*(shade[i3]),
						pts, 3, True, True);
				}
				if ((pm & (1 << invi1)) != 0 ||
						(pm & (16 << invi1)) != 0) {
					pts[0] = points[i];
					pts[1] = midPoint;
					pts[2] = edgePoints[i3];
					pts[3] = points[i];
					POLYGON(w, dr, *(shade[i1]),
						*(shade[i1]),
						pts, 3, True, True);
				}
			}
		}
	}
}

static void
drawTrapazoid(WelltrisWidget w, int cid,
		int ulx, int uly, int base_x, int base_y,
		int lrx, int lry, int plateau_x, int plateau_y)
{
	Point trapazoidList[MAX_SIDES + 1];
	GC gc;

	if (cid == CLEAR_COLOR)
		gc = w->welltris.inverseGC[2];
	else if (cid == BORDER_COLOR)
		gc = w->welltris.inverseGC[3];
	else
		gc = w->welltris.thingGC[cid][2];
	trapazoidList[0].x = ulx + w->welltris.offset.x;
	trapazoidList[0].y = uly + w->welltris.offset.y;
	trapazoidList[1].x = base_x;
	trapazoidList[1].y = base_y;
	trapazoidList[2].x = lrx - base_x - ulx;
	trapazoidList[2].y = lry - base_y - uly;
	trapazoidList[3].x = -plateau_x;
	trapazoidList[3].y = -plateau_y;
	trapazoidList[4].x = ulx + plateau_x - lrx;
	trapazoidList[4].y = uly + plateau_y - lry;
	POLYLINE((WelltrisWidget) w, dr,
		gc, trapazoidList, MAX_SIDES + 1, False);
}

static void
fillTrapazoidPosition(WelltrisWidget w,
		int pmid, int cid, int i, int j,
		Boolean graphicalXor)
{
	int ulx, uly, lrx, lry, base, plateau, baseWidth, endPt, wall;
	int newI = i;

	baseWidth = w->welltris.diameter * w->welltris.boxSize;
	endPt = 3 * baseWidth;
	wall = newI / w->welltris.diameter;
	newI -= (wall * w->welltris.diameter);
	ulx = (endPt + (2 * newI - w->welltris.diameter) *
		((w->welltris.depth / 2 +
		2 * w->welltris.boxSize * (w->welltris.depth - j)) /
		w->welltris.depth + w->welltris.boxSize)) / 2;
	uly = (j * baseWidth + w->welltris.depth / 2) / w->welltris.depth;
	lrx = (endPt + (2 + 2 * newI - w->welltris.diameter) *
		((w->welltris.depth / 2 +
		2 * w->welltris.boxSize * (w->welltris.depth - j - 1)) /
		w->welltris.depth + w->welltris.boxSize)) / 2;
	lry = ((j + 1) * baseWidth + w->welltris.depth / 2) /
		w->welltris.depth;
	base = w->welltris.boxSize + ((w->welltris.depth - j) *
		2 * w->welltris.boxSize + w->welltris.depth / 2) /
		w->welltris.depth;
	plateau = w->welltris.boxSize + ((w->welltris.depth - j - 1) * 2 *
		w->welltris.boxSize + w->welltris.depth / 2) /
		w->welltris.depth;
	switch (wall) {
	case 0:
		fillTrapazoid (w,
			pmid, cid,
			ulx + 1, uly, base, 0,
			lrx + 1, lry, plateau, 0,
			graphicalXor);
		break;
	case 1:
		fillTrapazoid (w,
			pmid, cid,
			endPt - uly + 1, ulx + 1, 0, base,
			endPt - lry + 1, lrx + 1, 0, plateau,
			graphicalXor);
			break;
	case 2:
		fillTrapazoid (w,
			pmid + BITMAPS, cid,
			endPt - ulx, endPt - uly + 1, -base, 0,
			endPt - lrx, endPt - lry + 1, -plateau, 0,
			graphicalXor);
			break;
	case 3:
		fillTrapazoid (w,
			pmid + BITMAPS, cid,
			uly, endPt - ulx, 0, -base,
			lry, endPt - lrx, 0, -plateau,
			graphicalXor);
			break;
	default:
		(void) fprintf (stderr,
			"trapazoid kinds range 0-3, got %d.\n", wall);
	}
}

static void
drawTrapazoidBorder(WelltrisWidget w, int cid, int i, int j)
{
	int ulx, uly, lrx, lry, base, plateau, baseWidth, endPt, wall;
	int newI = i;

	baseWidth = w->welltris.diameter * w->welltris.boxSize;
	endPt = 3 * baseWidth;
	wall = newI / w->welltris.diameter;
	newI -= (wall * w->welltris.diameter);
	ulx = (endPt + (2 * newI - w->welltris.diameter) *
		((w->welltris.depth / 2 +
		2 * w->welltris.boxSize * (w->welltris.depth - j)) /
		w->welltris.depth + w->welltris.boxSize)) / 2;
	uly = (j * baseWidth + w->welltris.depth / 2) / w->welltris.depth;
	lrx = (endPt + (2 + 2 * newI - w->welltris.diameter) *
		((w->welltris.depth / 2 +
		2 * w->welltris.boxSize * (w->welltris.depth - j - 1)) /
		w->welltris.depth + w->welltris.boxSize)) / 2;
	lry = ((j + 1) * baseWidth + w->welltris.depth / 2) /
		w->welltris.depth;
	base = w->welltris.boxSize + ((w->welltris.depth - j) *
		2 * w->welltris.boxSize + w->welltris.depth / 2) /
		w->welltris.depth;
	plateau = w->welltris.boxSize + ((w->welltris.depth - j - 1) * 2 *
		w->welltris.boxSize + w->welltris.depth / 2) /
		w->welltris.depth;
	if (!w->welltris.grid)
		cid = CLEAR_COLOR;
	switch (wall) {
	case 0:
		drawTrapazoid(w, cid,
			ulx + 1, uly, base, 0,
			lrx + 1, lry, plateau, 0);
		break;
	case 1:
		drawTrapazoid(w, cid,
			endPt - uly + 1, ulx + 1, 0, base,
			endPt - lry + 1, lrx + 1, 0, plateau);
			break;
	case 2:
		drawTrapazoid(w, cid,
			endPt - ulx, endPt - uly + 1, -base, 0,
			endPt - lrx, endPt - lry + 1, -plateau, 0);
			break;
	case 3:
		drawTrapazoid(w, cid,
			uly, endPt - ulx, 0, -base,
			lry, endPt - lrx, 0, -plateau);
			break;
	default:
		(void) fprintf (stderr,
			"trapazoid kinds range 0-3, got %d.\n", wall);
	}
}

static void
drawBox(WelltrisWidget w, int pmid, int cid, int i, int j)
{
	if (j < 0) {
		return;
	} else if (j < w->welltris.depth) {
		fillTrapazoidPosition(w,
			pmid, cid, i, j, False);
		drawTrapazoidBorder(w,
			BORDER_COLOR,
			i, j);
	} else {
		int base_i, base_j;

		wallToBase(w, &base_i, &base_j, i, j);
		drawUnit(w, pmid, cid, base_i, base_j);
#ifdef BOTTOM_BORDER
		drawUnitBorder(w, base_i, base_j);
#endif
	}
}

static void
drawField(WelltrisWidget w)
{
	int i, j, offset;

	for (j = 0; j < w->welltris.depth; j++) {
		offset = j * w->welltris.perimeter;
		for (i = 0; i < w->welltris.perimeter; i++) {
			if (w->welltris.wall[offset + i].pmid >= 0) {
				fillTrapazoidPosition(w,
					w->welltris.wall[offset + i].pmid,
					w->welltris.wall[offset + i].cid,
					i, j, False);
				drawTrapazoidBorder(w,
					BORDER_COLOR,
					i, j);
			}
			if (w->welltris.frozenWall[i / w->welltris.diameter] != 0) {
				fillTrapazoidPosition(w,
					w->welltris.wall[offset + i].pmid,
					((w->welltris.wall[offset + i].pmid >= 0) ?
					w->welltris.wall[offset + i].cid :
					CLEAR_COLOR),
					i, j, True);
				drawTrapazoidBorder(w,
					BORDER_COLOR,
					i, j);
			}
		}
	}
	for (j = 0; j < w->welltris.diameter; j++) {
		offset = j * w->welltris.diameter;
		for (i = 0; i < w->welltris.diameter; i++)
			if (w->welltris.base[offset + i].pmid >= 0) {
				drawUnit(w,
					w->welltris.base[offset + i].pmid,
					w->welltris.base[offset + i].cid,
					i, j);
#ifdef BOTTOM_BORDER
				drawUnitBorder(w, i, j);
#endif
			}
	}
}

static void
clearField(WelltrisWidget w)
{
	int i, j;

	for (j = 0; j < w->welltris.depth; j++)
		for (i = 0; i < w->welltris.perimeter; i++) {
			fillTrapazoidPosition(w,
				BITMAPS - 1, CLEAR_COLOR,
				i, j, False);
			drawTrapazoidBorder(w,
				BORDER_COLOR,
				i, j);
		}
	for (j = 0; j < w->welltris.diameter; j++)
		for (i = 0; i < w->welltris.diameter; i++) {
			clearUnit(w, i, j);
#ifdef BOTTOM_BORDER
			drawUnitBorder(w, i, j);
#endif
		}
}

static void
drawThing(WelltrisWidget w)
{
	int i, j, xi, yj;
	int *tempBase;

	if (!(tempBase = (int *) malloc(sizeof (int) *
			w->welltris.diameter * w->welltris.diameter))) {
		DISPLAY_ERROR("Not enough memory (Thing), exiting.");
	}
	/* Initialize */
	for (j = 0; j < w->welltris.diameter; j++) {
		yj = j * w->welltris.diameter;
		for (i = 0; i < w->welltris.diameter; i++)
			tempBase[yj + i] = EMPTY;
	}
	/* Find pieces and overlap them if necessary */
	for (i = 0; i < w->welltris.curThing.size; i++)
		for (j = 0; j < w->welltris.curThing.size; j++)
			if (tris[w->welltris.curThing.units].polyomino
					[w->welltris.curThing.polyNumber].shape[j][i] != 0) {
				xi = (w->welltris.curThing.xpos + i) % w->welltris.perimeter;
				yj = w->welltris.curThing.ypos + j;
				if (yj >= w->welltris.depth) {
					int base_i, base_j, base_ij;
					int rotPmid = rotateIndex(w,
						polyPixmap(i, j,
						(tris[w->welltris.curThing.units].polyomino
						[w->welltris.curThing.polyNumber].shape),
						w->welltris.curThing.size), xi, yj);

					wallToBase(w, &base_i, &base_j, xi, yj);
					xi = checkWall(w, rotPmid,
						base_i, base_j);
					base_ij = base_j * w->welltris.diameter + base_i;
					tempBase[base_ij] &= xi;
				}
			}
	/* Now draw */
	for (i = 0; i < w->welltris.curThing.size; i++)
		for (j = 0; j < w->welltris.curThing.size; j++)
			if (tris[w->welltris.curThing.units].polyomino
					[w->welltris.curThing.polyNumber].shape[j][i] != 0) {
				xi = (w->welltris.curThing.xpos + i) % w->welltris.perimeter;
				yj = w->welltris.curThing.ypos + j;
				if (yj >= w->welltris.depth) {
					int base_i, base_j, base_ij;

					wallToBase(w, &base_i, &base_j, xi, yj);
					base_ij = base_j * w->welltris.diameter + base_i;
					if (tempBase[base_ij] != EMPTY) {
						drawUnit(w,
							tempBase[base_ij],
							w->welltris.curThing.colorNumber,
							base_i, base_j);
#ifdef BOTTOM_BORDER
						drawUnitBorder(w, base_i, base_j);
#endif
						tempBase[base_ij] = EMPTY;
					}
				} else if (yj >= 0) {
					int rotPmid = rotateIndex(w,
						polyPixmap(i, j,
						(tris[w->welltris.curThing.units].polyomino
						[w->welltris.curThing.polyNumber].shape),
						w->welltris.curThing.size),
						xi, yj);

					fillTrapazoidPosition(w,
						checkWallSides(w, checkWallBottom(w,
						rotPmid, yj), xi),
						w->welltris.curThing.colorNumber,
						xi, yj, False);
					drawTrapazoidBorder(w,
						BORDER_COLOR,
						xi, yj);
				}
			}
	free(tempBase);
}

static void
drawThingDiff(WelltrisWidget w, Thing oldThing)
{
	int i, j, xi, yj, ox, oy;
	int base_i, base_j, base_ij;
	int *tempBase, *tmp;

	if (!(tempBase = (int *) malloc(sizeof (int) *
			w->welltris.diameter * w->welltris.diameter))) {
		DISPLAY_ERROR("Not enough memory (ThingDiff), exiting.");
	}
	/* Initialize */
	for (j = 0; j < w->welltris.diameter; j++) {
		base_j = j * w->welltris.diameter;
		for (i = 0; i < w->welltris.diameter; i++)
			tempBase[base_j + i] = EMPTY;
	}
	/* Find pieces and overlap them if necessary */
	for (i = 0; i < w->welltris.curThing.size; i++)
		for (j = 0; j < w->welltris.curThing.size; j++)
			if (tris[w->welltris.curThing.units].polyomino
			[w->welltris.curThing.polyNumber].shape[j][i] != 0) {
				xi = (w->welltris.curThing.xpos + i) % w->welltris.perimeter;
				yj = w->welltris.curThing.ypos + j;
				if (yj >= w->welltris.depth) {
					int rotPmid = rotateIndex(w, polyPixmap(i, j,
						(tris[w->welltris.curThing.units].polyomino
						[w->welltris.curThing.polyNumber].shape),
						w->welltris.curThing.size), xi, yj);

					wallToBase(w, &base_i, &base_j, xi, yj);
					xi = checkWall(w, rotPmid,
						base_i, base_j);
					base_ij = base_j * w->welltris.diameter + base_i;
					tempBase[base_ij] &= xi;
				}
			}
	/* Test to make corner split neighbor awareness */
	for (j = 0; j < w->welltris.diameter; j++) {
		base_j = j * w->welltris.diameter;
		for (i = 0; i < w->welltris.diameter; i++) {
			tmp = &tempBase[base_j + i];
			if (*tmp >= 0) {
				if ((j > 0) && (tempBase[base_j - w->welltris.diameter + i] != EMPTY)) {
					*tmp &= ~UP;
				}
				if ((j < w->welltris.diameter - 1) &&
						(tempBase[base_j + w->welltris.diameter + i] != EMPTY)) {
					*tmp &= ~DOWN;
				}
				if ((i > 0) && (tempBase[base_j + i - 1] != EMPTY && *tmp >= 0)) {
					*tmp &= ~LEFT;
				}
				if ((i < w->welltris.diameter - 1) &&
						(tempBase[base_j + i + 1] != EMPTY) && *tmp >= 0) {
					*tmp &= ~RIGHT;
				}
				if (i > 0 && j > 0 &&
						(tempBase[base_j - w->welltris.diameter + i - 1] != EMPTY)) {
					*tmp &= ~LEFT_UP;
				}
				if (i > 0 && (j < w->welltris.diameter - 1) &&
						(tempBase[base_j + w->welltris.diameter + i - 1] != EMPTY)) {
					*tmp &= ~LEFT_DOWN;
				}
				if ((i < w->welltris.diameter - 1) && j > 0 &&
						(tempBase[base_j - w->welltris.diameter + i + 1] != EMPTY)) {
					*tmp &= ~RIGHT_UP;
				}
				if ((i < w->welltris.diameter - 1) && (j < w->welltris.diameter - 1) &&
						(tempBase[base_j + w->welltris.diameter + i + 1] != EMPTY)) {
					*tmp &= ~RIGHT_DOWN;
				}
			}
		}
	}
	/* Now draw */
	for (i = 0; i < w->welltris.curThing.size; i++)
		for (j = 0; j < w->welltris.curThing.size; j++)
			if (tris[w->welltris.curThing.units].polyomino
			[w->welltris.curThing.polyNumber].shape[j][i] != 0) {
				xi = (w->welltris.curThing.xpos + i) % w->welltris.perimeter;
				yj = w->welltris.curThing.ypos + j;
				if (yj >= w->welltris.depth) {
					wallToBase(w, &base_i, &base_j, xi, yj);
					base_ij = base_j * w->welltris.diameter + base_i;
					if (tempBase[base_ij] > 0) {
						drawUnit(w, tempBase[base_ij],
						w->welltris.curThing.colorNumber, base_i, base_j);
#ifdef BOTTOM_BORDER
						drawUnitBorder(w, base_i, base_j);
#endif
						/* negate for erase */
						tempBase[base_ij] =
							-tempBase[base_ij];
					}
				} else if (yj >= 0) {
					int rotPmid = rotateIndex(w,
						polyPixmap(i, j,
						(tris[w->welltris.curThing.units].polyomino
						[w->welltris.curThing.polyNumber].shape),
						w->welltris.curThing.size), xi, yj);

					fillTrapazoidPosition(w,
						checkWallSides(w, checkWallBottom(w,
						rotPmid, yj), xi),
					w->welltris.curThing.colorNumber,
					xi, yj, False);
					drawTrapazoidBorder(w,
						BORDER_COLOR,
						xi, yj);
				}
		}
	/* Erase */
	for (i = 0; i < oldThing.size; i++)
		for (j = 0; j < oldThing.size; j++) {
			xi = (oldThing.xpos + i) % w->welltris.perimeter;
			yj = oldThing.ypos + j;
			ox = (xi - w->welltris.curThing.xpos + w->welltris.perimeter) % w->welltris.perimeter;
			oy = yj - w->welltris.curThing.ypos;
			if (yj >= w->welltris.depth) {
				wallToBase(w, &base_i, &base_j, xi, yj);
				if (tris[oldThing.units].polyomino
				[oldThing.polyNumber].shape[j][i] != 0 &&
				((ox >= w->welltris.curThing.size) ||
				(oy < 0) || (oy >= w->welltris.curThing.size) ||
				tempBase[base_j * w->welltris.diameter + base_i] == EMPTY)) {
					clearUnit(w, base_i, base_j);
#ifdef BOTTOM_BORDER
					drawUnitBorder(w, base_i, base_j);
#endif
				}
			} else if (yj >= 0 &&
			    (tris[oldThing.units].polyomino
			    [oldThing.polyNumber].shape[j][i] != 0) &&
			    ((ox >= w->welltris.curThing.size) ||
			    (oy < 0) || (oy >= w->welltris.curThing.size) ||
			    tris[w->welltris.curThing.units].polyomino
			    [w->welltris.curThing.polyNumber].shape[oy][ox] == 0)) {
				fillTrapazoidPosition(w,
					BITMAPS - 1, CLEAR_COLOR,
					xi, yj, False);
				drawTrapazoidBorder(w,
					BORDER_COLOR, xi, yj);
			}
		}
	free(tempBase);
}

static void
drawGrid(WelltrisWidget w)
{
	int i, j;

	for (j = 0; j < w->welltris.depth; j++)
		for (i = 0; i < w->welltris.perimeter; i++)
			drawTrapazoidBorder(w, BORDER_COLOR, i, j);
#ifdef BOTTOM_BORDER
	for (j = 0; j < w->welltris.diameter; j++)
		for (i = 0; i < w->welltris.diameter; i++)
			drawUnitBorder(w, i, j);
#endif
#if 0
	int i, k, baseWidth, square, endPt, newSum;
	int sum, sum_var;
	GC gc;

	if (w->welltris.grid)
		gc = w->welltris.inverseGC[3];
	else
		gc = w->welltris.inverseGC[2];
	baseWidth = w->welltris.diameter * w->welltris.boxSize;
	endPt = 3 * baseWidth;
	sum = sum_var = w->welltris.diameter * w->welltris.boxSize;
	square = 0;
	for (i = 0; i < w->welltris.diameter; i++) {
		DRAWLINE(w, dr, gc,
			square + w->welltris.offset.x,
			w->welltris.offset.y,
			sum_var + w->welltris.offset.x,
			sum + w->welltris.offset.y);
		DRAWLINE(w, dr, gc,
			endPt - 1 + w->welltris.offset.x,
			square + w->welltris.offset.y,
			endPt - 1 - sum + w->welltris.offset.x,
			sum_var + w->welltris.offset.y);
		DRAWLINE(w, dr, gc,
			endPt - 1 - square + w->welltris.offset.x,
			endPt - 1 + w->welltris.offset.y,
			endPt - 1 - sum_var + w->welltris.offset.x,
			endPt - 1 - sum + w->welltris.offset.y);
		DRAWLINE(w, dr, gc,
			w->welltris.offset.x,
			endPt - 1 - square + w->welltris.offset.y,
			sum + w->welltris.offset.x,
			endPt - 1 - sum_var + w->welltris.offset.y);
		square += 3 * w->welltris.boxSize;
		sum_var += w->welltris.boxSize;
	}
	newSum = endPt;
	square = 0;
	k = (w->welltris.diameter * w->welltris.boxSize) / w->welltris.depth;
	for (i = 0; i <= w->welltris.depth; i++) {
		FILLRECTANGLE(w, dr, gc,
			square + w->welltris.offset.x,
			square + w->welltris.offset.y,
			1, newSum);
		FILLRECTANGLE(w, dr, gc,
			square + w->welltris.offset.x,
			square + w->welltris.offset.y,
			newSum, 1);
		FILLRECTANGLE(w, dr, gc,
			square + newSum - 1 + w->welltris.offset.x,
			square + w->welltris.offset.y,
			1, newSum);
		FILLRECTANGLE(w, dr, gc,
			square + w->welltris.offset.x,
			square + newSum - 1 + w->welltris.offset.y,
			newSum, 1);
		square += k;
		newSum -= (2 * k);
	}
	square_x = square_y = baseSize;
	for (i = 0; i <= w->welltris.diameter ; i++) {
		FILLRECTANGLE(w, dr, gc,
			square_x, square_y, 1, baseSize);
		square_x += w->welltris.boxSize;
	}
	square_x = square_y = baseSize;
	for (i = 0; i <= w->welltris.diameter ; i++) {
		FILLRECTANGLE(w, dr, gc,
			square_x, square_y, baseSize, 1);
		square_y += w->welltris.boxSize;
	}
#endif
	FLUSH(w);
}

static void
draw3DFrame(const WelltrisWidget w, Boolean inside,
	int x, int y, int width, int height)
{
	GC gc;

	if (inside)
		gc = w->welltris.inverseGC[3];
	else
		gc = w->welltris.inverseGC[2];
	FILLRECTANGLE(w, dr, gc,
		x, y, width, 1);
	FILLRECTANGLE(w, dr, gc,
		x, y + 1, 1, height - 1);
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + 1, width - 2, 1);
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + 2, 1, height - 3);
	if (inside)
		gc = w->welltris.inverseGC[0];
	else
		gc = w->welltris.inverseGC[2];
	FILLRECTANGLE(w, dr, gc,
		x + 1, y + height - 1, width - 1, 1);
	FILLRECTANGLE(w, dr, gc,
		x + width - 1, y + 1, 1, height - 2);
	FILLRECTANGLE(w, dr, gc,
		x + 2, y + height - 2, width - 3, 1);
	FILLRECTANGLE(w, dr, gc,
		x + width - 2, y + 2, 1, height - 4);
}

static void
eraseFrame(WelltrisWidget w)
{
	FILLRECTANGLE(w, dr, w->welltris.inverseGC[2],
		0, 0, w->core.width, w->core.height);
}

static void
drawFrame(WelltrisWidget w, Boolean focus)
{
	int offsetX = w->welltris.offset.x - 2;
	int offsetY = w->welltris.offset.y - 2;
	int lengthX = w->welltris.offset.x + w->welltris.viewLength + 4;
	int lengthY = w->welltris.offset.y + w->welltris.viewLength + 4;

	FILLRECTANGLE(w, dr, w->welltris.inverseGC[4],
		0, 0, w->core.width, offsetY);
	FILLRECTANGLE(w, dr, w->welltris.inverseGC[4],
		0, offsetY, offsetX, lengthY - offsetY);
	FILLRECTANGLE(w, dr, w->welltris.inverseGC[4],
		lengthX, offsetY, w->core.width - lengthX, lengthY - offsetY);
	FILLRECTANGLE(w, dr, w->welltris.inverseGC[4],
		0, lengthY, w->core.width, offsetY + 1);
	draw3DFrame(w, focus,
		offsetX, offsetY, lengthX - offsetX, lengthY - offsetY);
}

static void
freezeWall(WelltrisWidget w, int wall, Boolean freeze)
{
	int i, j, offsetX, offsetY;

	offsetX = wall * w->welltris.diameter;
	for (j = 0; j < w->welltris.depth; j++) {
		offsetY = j * w->welltris.perimeter;
		for (i = 0; i < w->welltris.diameter; i++) {
			fillTrapazoidPosition(w,
				w->welltris.wall[offsetY + offsetX + i].pmid,
				((w->welltris.wall[offsetY + offsetX + i].pmid >= 0) ?
				w->welltris.wall[offsetY + offsetX + i].cid :
				CLEAR_COLOR),
				offsetX + i, j, freeze);
			drawTrapazoidBorder(w,
				BORDER_COLOR,
				offsetX + i, j);
		}
	}
#ifdef USE_SOUND
	if (w->welltris.sound) {
		playSound((char *) DRIPSOUND);
	}
#endif
	FLUSH(w);
}

static void
dropWall(WelltrisWidget w, int wall)
{
	int i, j, k, l, lines = 0, base_x, base_y;
	Boolean fits;

	for (j = w->welltris.depth - 1; j >= 0 && lines == 0; j--)
		for (i = 0; i < w->welltris.diameter; i++)
			if (w->welltris.wall[j * w->welltris.perimeter + wall * w->welltris.diameter + i].pmid != EMPTY)
				lines = w->welltris.depth - j;
	if (lines > 0) {
		fits = True;
		j = 0;
		while (j < (w->welltris.diameter / 2 + lines - 1) && fits) {
			j++;
			for (l = w->welltris.depth - j; l < w->welltris.depth; l++)
				for (i = 0; i < w->welltris.diameter; i++)
					if (w->welltris.wall[l * w->welltris.perimeter + wall * w->welltris.diameter + i].pmid != EMPTY) {
						wallToBase(w, &base_x, &base_y,
						wall * w->welltris.diameter + i, l + j);
						if (w->welltris.base[base_y * w->welltris.diameter + base_x].pmid != EMPTY)
							fits = False;
					}
		}
		if (!fits)
			j--;
		if (j > 0) {
			for (l = w->welltris.depth - 1; l >= 0; l--)
				for (i = 0; i < w->welltris.diameter; i++)
					if (w->welltris.wall[l * w->welltris.perimeter + wall * w->welltris.diameter + i].pmid != EMPTY) {
						k = wall * w->welltris.diameter + i;
						if (l + j >= w->welltris.depth) {
							wallToBase(w, &base_x, &base_y, k, l + j);
							w->welltris.base[base_y * w->welltris.diameter + base_x] =
							w->welltris.wall[l * w->welltris.perimeter + k];
							w->welltris.base[base_y * w->welltris.diameter + base_x].pmid =
							checkWall(w, rotateIndex(w,
							w->welltris.base[base_y * w->welltris.diameter + base_x].pmid,
							k, l + j),
							base_x, base_y);
							w->welltris.wall[l * w->welltris.perimeter + k].pmid = EMPTY;
							fillTrapazoidPosition(w,
								BITMAPS - 1,
								CLEAR_COLOR,
								k, l, False);
							drawTrapazoidBorder(w,
								BORDER_COLOR,
								k, l);
							drawUnit(w, w->welltris.base[base_y * w->welltris.diameter + base_x].pmid,
								w->welltris.base[base_y * w->welltris.diameter + base_x].cid,
								base_x, base_y);
#ifdef BOTTOM_BORDER
							drawUnitBorder(w, base_x, base_y);
#endif
						} else {
							w->welltris.wall[(l + j) * w->welltris.perimeter + k] = w->welltris.wall[l * w->welltris.perimeter + k];
							w->welltris.wall[(l + j) * w->welltris.perimeter + k].pmid =
							checkWallBottom(w, w->welltris.wall[(l + j) * w->welltris.perimeter + k].pmid, l + j);
							w->welltris.wall[l * w->welltris.perimeter + k].pmid = EMPTY;
							fillTrapazoidPosition(w,
								BITMAPS - 1,
								CLEAR_COLOR,
								k, l, False);
							drawTrapazoidBorder(w,
								BORDER_COLOR,
								k, l);
							fillTrapazoidPosition(w,
								w->welltris.wall[(l + j) * w->welltris.perimeter + k].pmid,
								w->welltris.wall[l * w->welltris.perimeter + k].cid,
								k, l + j, False);
							drawTrapazoidBorder(w,
								BORDER_COLOR,
								k, l + j);
						}
						FLUSH(w);
					}
#ifdef USE_SOUND
			if (w->welltris.sound) {
				playSound((char *) DRIPSOUND);
			}
#endif
			FLUSH(w);
		}
	}
}

static void
drawNext(WelltrisWidget w)
{
	setGameNext(w, ACTION_DRAW_NEXT,
		w->welltris.nextThing.randomNumber,
		w->welltris.nextThing.randomRotation,
		w->welltris.nextThing.randomReflection,
		w->welltris.units, w->welltris.diagonal, w->welltris.mixed,
		w->welltris.next.bonus);
}

static void
drawStatus(WelltrisWidget w)
{
	setGame(w, ACTION_STATUS);
}

void
redoNext(WelltrisWidget w)
{
	int diagonal = (w->welltris.next.diagonal) ? 1 : 0;
	int turnStyle = tris[w->welltris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	int next_start, i;

	w->welltris.nextThing.units = w->welltris.next.units - MIN_UNITS +
		((w->welltris.nextThing.bonus) ? 1 : 0);
	next_start = tris[w->welltris.nextThing.units].mode
		[diagonal].number[(w->welltris.nextThing.bonus) ?
		NOMIX : ((w->welltris.next.mixed) ? 1 : 0)][turnStyle];
	if (!next_start) {
		DISPLAY_ERROR("Error, no polyominoes for this combination");
	}
	next_start = w->welltris.nextThing.randomNumber % next_start;
	w->welltris.nextThing.colorNumber = next_start;

	if (w->welltris.next.mixed && !w->welltris.nextThing.bonus) {
		w->welltris.nextThing.units = 0;
		while (next_start >= tris[w->welltris.nextThing.units].mode
				[diagonal].number[NOMIX][turnStyle]) {
			next_start -= tris[w->welltris.nextThing.units].mode
				[diagonal].number[NOMIX][turnStyle];
			w->welltris.nextThing.units++;
		}
	}
	w->welltris.nextThing.polyNumber =
		tris[w->welltris.nextThing.units].mode
		[diagonal].start[next_start][turnStyle];
	if (tris[w->welltris.nextThing.units].mode[diagonal].turnStyle > NONE)
		for (i = 0; i < w->welltris.nextThing.randomRotation; i++)
			w->welltris.nextThing.polyNumber =
				tris[w->welltris.nextThing.units].polyomino
				[w->welltris.nextThing.polyNumber].rotation;
	if (tris[w->welltris.nextThing.units].mode[diagonal].turnStyle == ALL)
		for (i = 0; i < w->welltris.nextThing.randomReflection; i++)
			w->welltris.nextThing.polyNumber =
				tris[w->welltris.nextThing.units].polyomino
				[w->welltris.nextThing.polyNumber].reflection;
	w->welltris.nextThing.size =
		tris[w->welltris.nextThing.units].polyomino
		[w->welltris.nextThing.polyNumber].size;
	w->welltris.nextThing.ypos =
		-tris[w->welltris.nextThing.units].polyomino
		[w->welltris.nextThing.polyNumber].startHeight[(w->welltris.suddenAppear) ? JUMPIN : GRADUAL];
}

static void
initFill(WelltrisWidget w)
{
	int i, j, col, offset, fixedUnits;

	if (w->welltris.fixedFill) {
		/* This algorithm randomly picks fixedUnits per row
		 * (4 for DEFAULT_DIAMETER, ie. 8/2)
		 * This is more in keeping with the original */
		fixedUnits = w->welltris.diameter / 2;
		for (j = 0; j < w->welltris.depth; j++) {
			offset = j * w->welltris.perimeter;
			for (i = 0; i < w->welltris.perimeter; i++) {
				w->welltris.wall[offset + i].pmid = -1;
				w->welltris.wall[offset + i].cid = 0;
			}
			if (j >= w->welltris.depth - w->welltris.fill)
				for (i = 0; i < fixedUnits; i++) {
					do
						col = NRAND(w->welltris.perimeter);
					while (w->welltris.wall[offset + col].pmid != -1);
					w->welltris.wall[offset + col].pmid = BITMAPS - 1;
					w->welltris.wall[offset + col].cid = NRAND(w->welltris.numberPolyominoes);
				}
		}
	} else {
		/* This algorithm randomly fills boxes to a certain level
		 * The problem with this is that it could potentially fill
		 * up a whole row */
		for (j = 0; j < w->welltris.depth; j++) {
			offset = j * w->welltris.perimeter;
			for (i = 0; i < w->welltris.perimeter; i++)
				if ((j >= w->welltris.depth - w->welltris.fill) &&
						(NRAND(8) == 0)) {
					w->welltris.wall[offset + i].pmid = BITMAPS - 1;
					w->welltris.wall[offset + i].cid = NRAND(w->welltris.numberPolyominoes);
				} else {
					w->welltris.wall[offset + i].pmid = -1;
					w->welltris.wall[offset + i].cid = 0;
				}
		}
	}
	for (j = 0; j < w->welltris.diameter; j++) {
		offset = j * w->welltris.diameter;
		for (i = 0; i < w->welltris.diameter; i++) {
			w->welltris.base[offset + i].pmid = -1;
			w->welltris.base[offset + i].cid = 0;
		}
	}
}

static void
fillLines(WelltrisWidget w)
{
	int i, j;
	int diagonal = (w->welltris.current.diagonal) ? 1 : 0;
	int mixed = (w->welltris.current.mixed) ? 1 : 0;
	int turnStyle = tris[w->welltris.current.units -
		MIN_UNITS].mode[diagonal].turnStyle;

	FLUSH(w);
	for (j = 0; j < w->welltris.depth + w->welltris.diameter; j++) {
		for (i = 0; i < w->welltris.perimeter; i++) {
			drawBox(w, BITMAPS - 1,
				NRAND(tris[w->welltris.current.units -
				MIN_UNITS].mode[diagonal].number[mixed][turnStyle]), i, j);
		}
		FLUSH(w);
		Sleep(100);
	}
	Sleep(1000);
}

static Boolean
overlapping(WelltrisWidget w)
{
	int i, j;
	int x = w->welltris.curThing.xpos, y = w->welltris.curThing.ypos;

	for (i = 0; i < w->welltris.curThing.size; i++)
		for (j = 0; j < w->welltris.curThing.size; j++)
			if (tris[w->welltris.curThing.units].polyomino
			[w->welltris.curThing.polyNumber].shape[j][i] != 0) {
				int xi = (x + i) % w->welltris.perimeter;
				int yj = y + j;
				int offset = yj * w->welltris.perimeter;

				if (yj < w->welltris.depth) {
					if (w->welltris.frozenWall[xi / w->welltris.diameter] != 0)
						return True;
					if (w->welltris.suddenAppear) { /* creationalism */
/* This method does not allow turning polyomino to an area above screen */
						if ((yj < 0) || (w->welltris.wall[offset + xi].pmid >= 0))
							return True;
					} else { /* evolution */
/* This method one can turn polyomino to an area above of screen, also
 * part of the polyomino may not be visible initially */
						if ((yj >= 0) && (w->welltris.wall[offset + xi].pmid >= 0))
							return True;
					}
				} else if (yj < w->welltris.depth + w->welltris.diameter) {
					int base_i, base_j;

					wallToBase(w, &base_i, &base_j, xi, yj);
					if (w->welltris.base[base_j * w->welltris.diameter + base_i].pmid >= 0)
						return True;
				} else
					return True;
			}
	return False;
}

static Boolean
atBottom(WelltrisWidget w)
{
	int i, j;
	int x = w->welltris.curThing.xpos, y = w->welltris.curThing.ypos;

	for (i = 0; i < w->welltris.curThing.size; i++)
		for (j = 0; j < w->welltris.curThing.size; j++)
			if (tris[w->welltris.curThing.units].polyomino
			[w->welltris.curThing.polyNumber].shape[j][i] != 0) {
				int xi = (x + i) % w->welltris.perimeter;
				int yj = y + j + 1;
				int offset = yj * w->welltris.perimeter;

				if (yj < 0)
					 continue;
				if (yj < w->welltris.depth) {
					if (w->welltris.frozenWall[xi / w->welltris.diameter] != 0)
						return True;
					if (w->welltris.wall[offset + xi].pmid >= 0)
						return True;
				} else if (yj < w->welltris.depth + w->welltris.diameter) {
					int base_i, base_j;

					wallToBase(w, &base_i, &base_j, xi, yj);
					if (w->welltris.base[base_j * w->welltris.diameter + base_i].pmid >= 0)
						return True;
				} else
					return True;
			}
	return False;
}

static Boolean
atBaseFully(WelltrisWidget w)
{
	int i, j;

	for (i = 0; i < w->welltris.curThing.size; i++)
		for (j = 0; j < w->welltris.curThing.size; j++)
			if (w->welltris.curThing.ypos + j < w->welltris.depth &&
			tris[w->welltris.curThing.units].polyomino
			[w->welltris.curThing.polyNumber].shape[j][i] != 0)
				return False;
	return True;
}

static Boolean
atBasePartially(WelltrisWidget w)
{
	int i, j;

	for (i = 0; i < w->welltris.curThing.size; i++)
		for (j = 0; j < w->welltris.curThing.size; j++)
			if (w->welltris.curThing.ypos + j >= w->welltris.depth &&
			tris[w->welltris.curThing.units].polyomino
			[w->welltris.curThing.polyNumber].shape[j][i] != 0)
				return True;
	return False;
}

static Boolean
wallChange(WelltrisWidget w, Thing oldthing, Thing newthing)
{
	int wall = -1, i, j;
	int x = w->welltris.curThing.xpos;

	for (i = 0; i < oldthing.size; i++)
		for (j = 0; j < oldthing.size; j++)
			if (tris[oldthing.units].polyomino[oldthing.polyNumber].shape[j][i] != 0) {
				if (wall == -1)
					wall = ((x + i) % w->welltris.perimeter) / w->welltris.diameter;
				else if (wall != ((x + i) % w->welltris.perimeter) / w->welltris.diameter)
					return True;
			}
	for (i = 0; i < newthing.size; i++)
		for (j = 0; j < newthing.size; j++)
			if (tris[newthing.units].polyomino[newthing.polyNumber].shape[j][i] != 0)
				if (wall != ((x + i) % w->welltris.perimeter) / w->welltris.diameter)
					return True;
	return False;
}

static void
putBox(WelltrisWidget w)
{
	int i, j;
	int base_i, base_j, base_ij;
	int *tempBase, *tmp;

	if (!(tempBase = (int *) malloc(sizeof (int) *
			w->welltris.diameter * w->welltris.diameter))) {
		DISPLAY_ERROR("Not enough memory (putBox), exiting.");
	}
	/* Initialize */
	for (j = 0; j < w->welltris.diameter; j++) {
		base_j = j * w->welltris.diameter;
		for (i = 0; i < w->welltris.diameter; i++)
			tempBase[base_j + i] = EMPTY;
	}
	/* Set pieces and overlap them if necessary */
	for (i = 0; i < w->welltris.curThing.size; i++)
		for (j = 0; j < w->welltris.curThing.size; j++)
			if (tris[w->welltris.curThing.units].polyomino
			[w->welltris.curThing.polyNumber].shape[j][i] != 0) {
				int xi = (w->welltris.curThing.xpos + i) % w->welltris.perimeter;
				int yj = w->welltris.curThing.ypos + j;

				if (yj >= w->welltris.depth) {
					int rotPmid = rotateIndex(w, polyPixmap(i, j,
					(tris[w->welltris.curThing.units].polyomino
					[w->welltris.curThing.polyNumber].shape),
					w->welltris.curThing.size), xi, yj);

					wallToBase(w, &base_i, &base_j, xi, yj);
					xi = checkWall(w, rotPmid,
						base_i, base_j);
					base_ij = base_j * w->welltris.diameter + base_i;
					tempBase[base_ij] &= xi;
				} else if (yj >= 0) {
					int rotPmid = rotateIndex(w, polyPixmap(i, j,
					(tris[w->welltris.curThing.units].polyomino
					[w->welltris.curThing.polyNumber].shape),
					w->welltris.curThing.size), xi, yj);

					w->welltris.wall[yj * w->welltris.perimeter + xi].pmid =
						checkWallSides(w, checkWallBottom(w,
						rotPmid, yj), xi),
					w->welltris.wall[yj * w->welltris.perimeter + xi].cid =
						w->welltris.curThing.colorNumber;
					w->welltris.frozenWall[xi / w->welltris.diameter] = MAX_SIDES;
				}
			}
	/* Test to make corner split neighbor awareness */
	for (j = 0; j < w->welltris.diameter; j++) {
		base_j = j * w->welltris.diameter;
		for (i = 0; i < w->welltris.diameter; i++) {
			tmp = &tempBase[base_j + i];
			if (*tmp >= 0) {
				if (j > 0) {
					if (tempBase[base_j - w->welltris.diameter + i] != EMPTY)
						*tmp &= ~UP;
				}
				if (j < w->welltris.diameter - 1) {
					if (tempBase[base_j + w->welltris.diameter + i] != EMPTY)
						*tmp &= ~DOWN;
				}
				if (i > 0) {
					if (tempBase[base_j + i - 1] != EMPTY && *tmp >= 0)
						*tmp &= ~LEFT;
				}
				if (i < w->welltris.diameter - 1) {
					if (tempBase[base_j + i + 1] != EMPTY && *tmp >= 0)
						*tmp &= ~RIGHT;
				}
				if (i > 0 && j > 0) {
					if (tempBase[base_j - w->welltris.diameter + i - 1] != EMPTY)
						*tmp &= ~LEFT_UP;
					}
				if (i > 0 && j < w->welltris.diameter - 1) {
					if (tempBase[base_j + w->welltris.diameter + i - 1] != EMPTY)
						*tmp &= ~LEFT_DOWN;
				}
				if (i < w->welltris.diameter - 1 && j > 0) {
					if (tempBase[base_j - w->welltris.diameter + i + 1] != EMPTY)
						*tmp &= ~RIGHT_UP;
				}
				if (i < w->welltris.diameter - 1 && j < w->welltris.diameter - 1) {
					if (tempBase[base_j + w->welltris.diameter + i + 1] != EMPTY)
						*tmp &= ~RIGHT_DOWN;
				}
			}
		}
	}
	for (j = 0; j < w->welltris.diameter; j++) {
		base_j = j * w->welltris.diameter;
		for (i = 0; i < w->welltris.diameter; i++) {
			base_ij = base_j + i;
			if (tempBase[base_ij] != EMPTY) {
				w->welltris.base[base_ij].pmid = tempBase[base_ij];
				w->welltris.base[base_ij].cid = w->welltris.curThing.colorNumber;
			}
		}
	}
	free(tempBase);
}

static int
checkLines(WelltrisWidget w)
{
	int *lSet[2], nSet[2];
	int i, j, k, *tmp, offset;

	if (!(lSet[0] = (int *) malloc(sizeof (int) * w->welltris.diameter))) {
		DISPLAY_ERROR("Not enough memory (lSet0), exiting.");
	}
	if (!(lSet[1] = (int *) malloc(sizeof (int) * w->welltris.diameter))) {
		DISPLAY_ERROR("Not enough memory (lSet1), exiting.");
	}
	nSet[0] = nSet[1] = 0;

	/* Find filled rows */
	for (j = 0; j < w->welltris.diameter; j++) {
		lSet[0][j] = 0;
		offset = j * w->welltris.diameter;
		for (i = 0; i < w->welltris.diameter; i++)
			if (w->welltris.base[offset + i].pmid >= 0)
				lSet[0][j]++;
		if (lSet[0][j] == w->welltris.diameter)
			nSet[0]++;
	}
	/* Find filled columns */
	for (i = 0; i < w->welltris.diameter; i++) {
		lSet[1][i] = 0;
		for (j = 0; j < w->welltris.diameter; j++)
			if (w->welltris.base[j * w->welltris.diameter + i].pmid >= 0)
				lSet[1][i]++;
		if (lSet[1][i] == w->welltris.diameter)
			nSet[1]++;
	}

	if ((nSet[0] != 0) || (nSet[1] != 0)) {
		Field *tempBase;

		if (!(tempBase = (Field *) malloc(sizeof (Field) * w->welltris.diameter * w->welltris.diameter))) {
			DISPLAY_ERROR("Not enough memory (Field), exiting.");
		}

		for (k = 0; k < ((NUM_FLASHES / (nSet[0] + nSet[1])) / 2) * 2; k++) {
			for (j = 0; j < w->welltris.diameter; j++) {
				if (lSet[0][j] == w->welltris.diameter)
					for (i = 0; i < w->welltris.diameter; i++)
						xorUnit(w, i, j);
				if (lSet[1][j] == w->welltris.diameter)
					for (i = 0; i < w->welltris.diameter; i++)
						xorUnit(w, j, i);
			}
			FLUSH(w);
			Sleep(100);
		}

		for (j = 0; j < w->welltris.diameter; j++) {
			offset = j * w->welltris.diameter;
			for (i = 0; i < w->welltris.diameter; i++)
				tempBase[offset + i] = w->welltris.base[offset +i];
		}
		/* Shrink */
		if (nSet[0]) {
		  for (j = 0; j < w->welltris.diameter / 2; j++)
		    if (lSet[0][j] == w->welltris.diameter)
		      for (i = 0; i < w->welltris.diameter; i++) {
			for (k = j; k > 0; k--)
			  tempBase[k * w->welltris.diameter + i] =
				tempBase[(k - 1) * w->welltris.diameter + i];
			tempBase[i].pmid = EMPTY;
			if (j > 0) {
			  tmp = &tempBase[j * w->welltris.diameter + i].pmid;
			  if (*tmp >= 0)
			    *tmp = CHECKDOWN((unsigned int) *tmp);
			}
			tmp = &tempBase[(j + 1) * w->welltris.diameter + i].pmid;
			if (*tmp >= 0)
			  *tmp = CHECKUP((unsigned int) *tmp);
		      }
		  for (j = w->welltris.diameter - 1; j >= w->welltris.diameter / 2; j--)
		    if (lSet[0][j] == w->welltris.diameter)
		      for (i = 0; i < w->welltris.diameter; i++) {
			for (k = j; k < w->welltris.diameter - 1; k++)
			  tempBase[k * w->welltris.diameter + i] = tempBase[(k + 1 ) * w->welltris.diameter + i];
			tempBase[(w->welltris.diameter - 1) * w->welltris.diameter + i].pmid = EMPTY;
			if (j < w->welltris.diameter - 1) {
			  tmp = &tempBase[j * w->welltris.diameter + i].pmid;
			  if (*tmp >= 0)
			    *tmp = CHECKUP((unsigned int) *tmp);
			}
			tmp = &tempBase[(j - 1) * w->welltris.diameter + i].pmid;
			if (*tmp >= 0)
			  *tmp = CHECKDOWN((unsigned int) *tmp);
		    }
		}
		if (nSet[1]) {
		  for (i = 0; i < w->welltris.diameter / 2; i++)
		    if (lSet[1][i] == w->welltris.diameter)
		      for (j = 0; j < w->welltris.diameter; j++) {
			for (k = i; k > 0; k--)
			  tempBase[j * w->welltris.diameter + k] = tempBase[j * w->welltris.diameter + k - 1];
			tempBase[j * w->welltris.diameter].pmid = EMPTY;
			if (i > 0) {
			  tmp = &tempBase[j * w->welltris.diameter + i].pmid;
			  if (*tmp >= 0)
			    *tmp = CHECKRIGHT((unsigned int) *tmp);
			}
			tmp = &tempBase[j * w->welltris.diameter + i + 1].pmid;
			if (*tmp >= 0)
			  *tmp = CHECKLEFT((unsigned int) *tmp);
		      }
		  for (i = w->welltris.diameter - 1; i >= w->welltris.diameter / 2; i--)
		    if (lSet[1][i] == w->welltris.diameter)
		      for (j = 0; j < w->welltris.diameter; j++) {
			for (k = i; k < w->welltris.diameter - 1; k++)
			  tempBase[j * w->welltris.diameter + k] = tempBase[j * w->welltris.diameter + k + 1];
			tempBase[j * w->welltris.diameter + w->welltris.diameter - 1].pmid = EMPTY;
			if (i < w->welltris.diameter - 1) {
			  tmp = &tempBase[j * w->welltris.diameter + i].pmid;
			  if (*tmp >= 0)
			    *tmp = CHECKLEFT((unsigned int) *tmp);
			}
			tmp = &tempBase[j * w->welltris.diameter + i - 1].pmid;
			if (*tmp >= 0)
			  *tmp = CHECKRIGHT((unsigned int) *tmp);
		      }
		}
		for (j = 0; j < w->welltris.diameter; j++) {
			offset = j * w->welltris.diameter;
			for (i = 0; i < w->welltris.diameter; i++)
				if ((tempBase[j * w->welltris.diameter + i].cid != w->welltris.base[j * w->welltris.diameter + i].cid) ||
				(tempBase[j * w->welltris.diameter + i].pmid != w->welltris.base[j * w->welltris.diameter + i].pmid))
				{
				  w->welltris.base[j * w->welltris.diameter + i] = tempBase[j * w->welltris.diameter + i];
				  if (w->welltris.base[j * w->welltris.diameter + i].pmid >= 0)
				    drawUnit(w, w->welltris.base[j * w->welltris.diameter + i].pmid, w->welltris.base[j * w->welltris.diameter + i].cid, i, j);
				   else
				    clearUnit(w, i, j);
#ifdef BOTTOM_BORDER
				  drawUnitBorder(w, i, j);
#endif
				}
		}
		free(tempBase);
		FLUSH(w);
#ifdef USE_SOUND
		if (w->welltris.sound) {
			playSound((char *) BUMPSOUND);
		}
#endif
	}
	free(lSet[0]);
	free(lSet[1]);
	return (nSet[0] + nSet[1]);
}

static Boolean
allFrozen(WelltrisWidget w)
{
	int wall;

	for (wall = 0; wall < MAX_SIDES; wall++) {
		if (w->welltris.frozenWall[wall] == MAX_SIDES) {
			freezeWall(w, wall, True);
			FLUSH(w);
		}
	}
	for (wall = 0; wall < MAX_SIDES; wall++) {
		if (w->welltris.frozenWall[wall] == 0) {
			return False;
		}
	}
	return True;
}


static void
checkFreeze(WelltrisWidget w)
{
	int wall;

	/* randomize the subtleties */
	if (NRAND(2) == 1) {
		for (wall = 0; wall < MAX_SIDES; wall++) {
			if (w->welltris.frozenWall[wall] == 0)
				dropWall(w, wall);
			else {
				w->welltris.frozenWall[wall]--;
				if (w->welltris.frozenWall[wall] == 0) {
					freezeWall(w, wall, False);
					dropWall(w, wall);
				}
			}
		}
	} else {
		for (wall = MAX_SIDES - 1; wall >= 0; wall--) {
			if (w->welltris.frozenWall[wall] == 0)
				dropWall(w, wall);
			else {
				w->welltris.frozenWall[wall]--;
				if (w->welltris.frozenWall[wall] == 0) {
					freezeWall(w, wall, False);
					dropWall(w, wall);
				}
			}
		}
	}
}

static void
newThing(WelltrisWidget w)
{
	int wall;

	w->welltris.curThing = w->welltris.nextThing;
	/*w->welltris.current = w->welltris.next;*/
	w->welltris.current.units = w->welltris.next.units;
	w->welltris.current.diagonal = w->welltris.next.diagonal;
	w->welltris.current.mixed = w->welltris.next.mixed;
	w->welltris.current.bonus = w->welltris.next.bonus;
	w->welltris.nextThing.randomNumber = LRAND();
	w->welltris.nextThing.randomRotation = NRAND(MAX_SIDES);
	w->welltris.nextThing.randomReflection = NRAND(2);
	w->welltris.nextThing.bonus = (w->welltris.next.bonus) ? 1 : 0;
	/* not all frozen else this is an infinite loop */
	/* curThing.xpos = (w->welltris.diameter - curThing.size) / 2; */
	w->welltris.curThing.xpos = NRAND(w->welltris.diameter -
		w->welltris.curThing.size + 1);
	do {
		wall = NRAND(MAX_SIDES);
	} while (w->welltris.frozenWall[wall] != 0);
	w->welltris.curThing.xpos += wall * w->welltris.diameter;
	redoNext(w);
}

static void
clearThings(WelltrisWidget w, Boolean draw)
{
	int wall;

	for (wall = 0; wall < MAX_SIDES; wall++)
		w->welltris.frozenWall[wall] = 0;
	if (draw)
		clearField(w);
	initFill(w);
	if (draw) {
		if (w->welltris.grid)
			drawGrid(w);
		drawField(w);
	}
	w->welltris.delay = delay[w->welltris.level - 1];
	w->welltris.score = w->welltris.fill * w->welltris.level * 10;
	w->welltris.started = True;
	w->welltris.paused = False;
	w->welltris.pauseStart = 0;
	w->welltris.completedRows = 0;
	w->welltris.newRows = 0;
	drawStatus(w);
}

static void
gameOver(WelltrisWidget w, Boolean fast)
{
	if (!w->welltris.started)
		return;
	if (!fast)
		fillLines(w);
	setGame(w, ACTION_GAME_OVER);
	clearThings(w, True);
}

static Boolean
tryMove(WelltrisWidget w, move_t move)
{
	int diagonal = (w->welltris.current.diagonal) ? 1 : 0;
	Thing old;
	int i;
	Boolean skip = False;

	old = w->welltris.curThing;

	switch (move) {
	case FALL:
		w->welltris.curThing.ypos++;
		break;

	case DROP:
		do {
			w->welltris.curThing.ypos++;
			if (w->welltris.dropPaused == 0)
				w->welltris.score += w->welltris.level +
					!w->welltris.grid + w->welltris.fill;
		} while (!overlapping(w));
		w->welltris.curThing.ypos--;
		break;

	case MOVE_LEFT:
		w->welltris.curThing.xpos = (w->welltris.curThing.xpos +
			 w->welltris.perimeter - 1) % w->welltris.perimeter;
		if (atBaseFully(w) || (atBasePartially(w) &&
				wallChange(w, old, w->welltris.curThing)))
			skip = True;
		break;

	case MOVE_RIGHT:
		w->welltris.curThing.xpos = (w->welltris.curThing.xpos +
			 w->welltris.perimeter + 1) % w->welltris.perimeter;
		if (atBaseFully(w) || (atBasePartially(w) &&
				wallChange(w, old, w->welltris.curThing)))
			skip = True;
		break;

	case ROTATE:
		if (tris[w->welltris.current.units -
				MIN_UNITS].mode[diagonal].turnStyle > NONE) {
			if (w->welltris.cw)
				for (i = 0; i < MAX_SIDES - 1; i++)
					w->welltris.curThing.polyNumber =
						tris[w->welltris.curThing.units].polyomino
						[w->welltris.curThing.polyNumber].rotation;
			else
				w->welltris.curThing.polyNumber =
					tris[w->welltris.curThing.units].polyomino
					[w->welltris.curThing.polyNumber].rotation;
			w->welltris.curThing.xpos = old.xpos;
			w->welltris.curThing.ypos = old.ypos;
			if (atBasePartially(w) &&
					wallChange(w, old, w->welltris.curThing))
				skip = True;
		}
		break;

	case REFLECT: /* reflect on y axis */
		if (tris[w->welltris.current.units -
				MIN_UNITS].mode[diagonal].turnStyle == ALL) {
			w->welltris.curThing.polyNumber =
				tris[w->welltris.curThing.units].polyomino
				[w->welltris.curThing.polyNumber].reflection;
			w->welltris.curThing.xpos = old.xpos;
			w->welltris.curThing.ypos = old.ypos;
			if (atBasePartially(w) &&
					wallChange(w, old, w->welltris.curThing))
				skip = True;
		}
		break;
	}

	if (skip || overlapping(w)) {
		w->welltris.curThing = old;
		if (w->welltris.curThing.ypos < 0)
			return False;
	} else {
		drawThingDiff(w, old);
	}
	return True;
}

static Boolean
moveOne(WelltrisWidget w, move_t move)
{
	int lines;

	if ((move == DROP) || ((move == FALL) && atBottom(w))) {
		if (!tryMove(w, move)) {
			gameOver(w, False);
			return True;
		}
		putBox(w);
		lines = checkLines(w);
		if (w->welltris.dropPaused != 0) {
			w->welltris.dropPaused = 0;
		} else {
			w->welltris.score += (!w->welltris.grid + w->welltris.fill +
				w->welltris.level + lines + w->welltris.suddenAppear) *
				(w->welltris.current.bonus ? 1 : 3) *
				(w->welltris.showNext ? 1 : 3) /
				(1 + w->welltris.current.mixed);
			w->welltris.score += (w->welltris.showNext ? 3 : 5) *
				w->welltris.level * w->welltris.level * lines * lines *
				(w->welltris.current.bonus ? 3 : 5) *
				w->welltris.startLevel *
				(w->welltris.current.units + w->welltris.current.diagonal) *
				(w->welltris.current.units + w->welltris.current.diagonal) /
				(16 * (1 + w->welltris.current.mixed));
		}
		w->welltris.completedRows += lines;
		w->welltris.newRows += lines;
		w->welltris.next.bonus = False;
		if ((w->welltris.newRows > THRESHOLD(w->welltris.level)) &&
				(w->welltris.level < MAX_LEVELS)) {
			w->welltris.level++;
			w->welltris.newRows = 0;
			if (LEGAL(w->welltris.next.units,
					w->welltris.next.diagonal,
					w->welltris.next.mixed,
					w->welltris.bonus))
				w->welltris.next.bonus = True;
		}
		drawStatus(w);
		FLUSH(w);
		if (allFrozen(w)) {
			gameOver(w, False);
			return True;
		}
		newThing(w);
		drawThing(w);
		drawNext(w);
		FLUSH(w);
		checkFreeze(w);
		if (overlapping(w)) {
			gameOver(w, False);
		}
		return True;
	} else {
		if (w->welltris.dropPaused == 1)
			w->welltris.dropPaused = 2;
		(void) tryMove(w, move);
		if ((w->welltris.newRows > THRESHOLD(w->welltris.level)) &&
				(w->welltris.level < MAX_LEVELS)) {
			w->welltris.next.bonus = False;
			w->welltris.level++;
			w->welltris.newRows = 0;
			if (LEGAL(w->welltris.next.units,
					w->welltris.next.diagonal,
					w->welltris.next.mixed,
					w->welltris.bonus))
				w->welltris.next.bonus = True;
			drawStatus(w);
			drawNext(w);
		}
		return False;
	}
}

#ifndef WINVER
static
#endif
void
falldownGame(WelltrisWidget w)
{
	if (!w->welltris.paused) {
		w->welltris.gotNew = moveOne(w, FALL) || w->welltris.gotNew;
	} else {
		if (w->welltris.pauseStart >= 32) {
			clearThings(w, True);
		}
		if (!w->welltris.started)
			w->welltris.pauseStart++;
	}
}

#ifndef WINVER
static
#endif
void
moveGameInput(WelltrisWidget w, int direction)
{
	if (!w->welltris.paused)
		w->welltris.gotNew = moveOne(w, (move_t) direction) ||
			w->welltris.gotNew;
}

static void
resetThings(WelltrisWidget w)
{
	if (w->welltris.wall)
		free(w->welltris.wall);
	if (w->welltris.base)
		free(w->welltris.base);
	w->welltris.depth = 3 * (w->welltris.diameter / 2);
	w->welltris.perimeter = 4 * w->welltris.diameter;
	if (!(w->welltris.wall = (Field *) malloc(sizeof(Field) *
			(w->welltris.depth + w->welltris.diameter) *
			w->welltris.perimeter))) {
		DISPLAY_ERROR("Not enough memory (wall), exiting.");
	}
	if (!(w->welltris.base = (Field *) malloc(sizeof(Field) *
			w->welltris.diameter * w->welltris.diameter))) {
		DISPLAY_ERROR("Not enough memory (base), exiting.");
	}
	w->welltris.started = False;
}

static void
resizeThings(WelltrisWidget w)
{
#ifdef IMAGE
#ifndef WINVER
	Display *display = XtDisplay(w);
	Window window = XtWindow(w);
	XWindowAttributes xgwa;

	(void) XGetWindowAttributes(display, window, &xgwa);
	graypix = XCreateBitmapFromData(display, window,
		(char *) gray1_bits, 2, 2);
#endif
	free_images(FALSE);
	if (w->welltris.boxSize > 4)
		(void) create_images(
#ifndef WINVER
			display, xgwa.visual,
#endif
			(w->welltris.plainPiece) ? 0 : 1, FALSE,
			w->welltris.boxSize);
#endif
}

#define BRIGHT_FACTOR 0.8
#define DARK_FACTOR 0.75

#ifdef WINVER
#define MAX_INTENSITY 0xFF

int
brighter(const int light)
{
	int i = (int) ((1 - BRIGHT_FACTOR) * MAX_INTENSITY);
	int temp = light;

	if (temp < i)
		temp = i;
	return MIN(temp / BRIGHT_FACTOR, MAX_INTENSITY);
}

int
darker(const int light)
{
	return (int) (light * DARK_FACTOR);
}

static void
setValuesGame(WelltrisWidget w)
{
	struct tagColor {
		int red, green, blue;
	} color;
	char szBuf[80], buf[20];
	int thing;

	w->welltris.level = GetPrivateProfileInt(SECTION,
		"level", DEFAULT_LEVELS, INIFILE);
	w->welltris.units = GetPrivateProfileInt(SECTION,
		"units", DEFAULT_UNITS, INIFILE);
	w->welltris.diameter = GetPrivateProfileInt(SECTION,
		"diameter", DEFAULT_DIAMETER, INIFILE);
	w->welltris.fill = GetPrivateProfileInt(SECTION,
		"fill", 0, INIFILE);
	w->welltris.fixedFill = (BOOL) GetPrivateProfileInt(SECTION,
		"fixedFill", TRUE, INIFILE);
	w->welltris.suddenAppear = (BOOL) GetPrivateProfileInt(SECTION,
		"suddenAppear", FALSE, INIFILE);
	w->welltris.diagonal = (BOOL) GetPrivateProfileInt(SECTION,
		"diagonal", DEFAULT_DIAGONAL, INIFILE);
	w->welltris.mixed = (BOOL) GetPrivateProfileInt(SECTION,
		"mixed", DEFAULT_MIXED, INIFILE);
	w->welltris.bonus = (BOOL) GetPrivateProfileInt(SECTION,
		"bonus", FALSE, INIFILE);
	w->welltris.cw = (BOOL) GetPrivateProfileInt(SECTION,
		"cw", TRUE, INIFILE);
	w->welltris.plainPiece = (BOOL) GetPrivateProfileInt(SECTION,
		"plainPiece", FALSE, INIFILE);
	w->welltris.showNext = (BOOL) GetPrivateProfileInt(SECTION,
		"showNext", FALSE, INIFILE);
	w->welltris.grid = (BOOL) GetPrivateProfileInt(SECTION,
		"grid", TRUE, INIFILE);
	w->welltris.sound = (BOOL) GetPrivateProfileInt(SECTION,
		"sound", FALSE, INIFILE);
	(void) GetPrivateProfileString(SECTION,
		"bumpSound", BUMPSOUND, szBuf, sizeof (szBuf), INIFILE);
	(void) GetPrivateProfileString(SECTION,
		"dripSound", DRIPSOUND, szBuf, sizeof (szBuf), INIFILE);
	(void) strncpy(w->welltris.bumpSound, szBuf, 80);
	w->welltris.bumpSound[80] = 0;
	w->welltris.mono = (BOOL) GetPrivateProfileInt(SECTION,
		"mono", DEFAULT_MONO, INIFILE);
	w->welltris.reverse = (BOOL) GetPrivateProfileInt(SECTION,
		"reverseVideo", DEFAULT_REVERSE, INIFILE);
	/* #AEB2C3 */
	(void) GetPrivateProfileString(SECTION,
		"background", "255 255 255", szBuf, sizeof (szBuf), INIFILE);
	(void) sscanf(szBuf, "%d %d %d",
		&(color.red), &(color.green), &(color.blue));
	w->welltris.inverseGC[1] = RGB(color.red, color.green, color.blue);
	w->welltris.inverseGC[0] = RGB(brighter(color.red),
		brighter(color.green), brighter(color.blue));
	w->welltris.inverseGC[2] = RGB(darker(color.red),
		darker(color.green), darker(color.blue));
	w->welltris.inverseGC[3] = RGB(darker(darker(color.red)),
		darker(darker(color.green)), darker(darker(color.blue)));
	w->welltris.inverseGC[4] = RGB(darker(darker(darker(color.red))),
		darker(darker(darker(color.green))),
		darker(darker(darker(color.blue))));
	for (thing = 0; thing < MAX_START_POLYOMINOES; thing++) {
		(void) sprintf(buf, "thingColor%d", thing);
		(void) GetPrivateProfileString(SECTION,
			buf, polyominoColorString[thing],
			szBuf, sizeof (szBuf), INIFILE);
		(void) sscanf(szBuf, "%d %d %d",
			&(color.red), &(color.green), &(color.blue));
		w->welltris.thingGC[thing][2] =
			RGB(color.red, color.green, color.blue);
		w->welltris.thingGC[thing][1] =
			RGB(brighter(color.red), brighter(color.green),
			brighter(color.blue));
		w->welltris.thingGC[thing][3] =
			RGB(darker(color.red), darker(color.green),
			darker(color.blue));
		w->welltris.thingGC[thing][0] =
			RGB(brighter(brighter(color.red)),
			brighter(brighter(color.green)),
			brighter(brighter(color.blue)));
		w->welltris.thingGC[thing][4] =
			RGB(darker(darker(color.red)),
			darker(darker(color.green)),
			darker(darker(color.blue)));
	}
	(void) GetPrivateProfileString(SECTION,
		"userName", "Guest", szBuf, sizeof (szBuf), INIFILE);
	(void) strncpy(w->welltris.userName, szBuf, 80);
	(void) GetPrivateProfileString(SECTION,
		"scoreFile", "", szBuf, sizeof (szBuf), INIFILE);
	(void) strncpy(w->welltris.scoreFile, szBuf, 80);
}

void
destroyGame(HBRUSH brush)
{
	(void) DeleteObject(brush);
	PostQuitMessage(0);
}

#else
#define MAX_INTENSITY 0xFFFF

Pixel
brighter(WelltrisWidget w, Pixel pixel)
{
	XColor color;
	Colormap colormap = DefaultColormapOfScreen(XtScreen(w));
	int i = (int) ((1 - BRIGHT_FACTOR) * MAX_INTENSITY);

	color.pixel = pixel;
	VOID XQueryColor(XtDisplay(w), colormap, &color);
	if (color.red < i)
		color.red = i;
	if (color.green < i)
		color.green = i;
	if (color.blue < i)
		color.blue = i;
	color.red = (unsigned short) MIN(color.red / BRIGHT_FACTOR, MAX_INTENSITY);
	color.green = (unsigned short) MIN(color.green / BRIGHT_FACTOR, MAX_INTENSITY);
	color.blue = (unsigned short) MIN(color.blue / BRIGHT_FACTOR, MAX_INTENSITY);
	if (XAllocColor(XtDisplay(w), colormap, &color))
		return color.pixel;
	(void) fprintf(stderr,
		"Welltris: Could not allocate brighter color %ld: %d %d %d\n",
		color.pixel, color.red, color.green, color.blue);
	return pixel;
}

Pixel
darker(WelltrisWidget w, Pixel pixel)
{
	XColor color;
	Colormap colormap = DefaultColormapOfScreen(XtScreen(w));

	color.pixel = pixel;
	VOID XQueryColor(XtDisplay(w), colormap, &color);
	color.red = (unsigned short) (color.red * DARK_FACTOR);
	color.green = (unsigned short) (color.green * DARK_FACTOR);
	color.blue = (unsigned short) (color.blue * DARK_FACTOR);
	if (XAllocColor(XtDisplay(w), colormap, &color))
		return color.pixel;
	(void) fprintf(stderr,
		"Welltris: Could not allocate darker color %ld: %d %d %d\n",
		color.pixel, color.red, color.green, color.blue);
	return pixel;
}

static void
getColor(WelltrisWidget w, int thing)
{
	XGCValues values;
	XtGCMask valueMask;
	XColor colorCell, rgb;
	Boolean gotColor = False;

	valueMask = GCForeground | GCBackground;
	if (w->welltris.reverse) {
		values.background = w->welltris.foreground;
	} else {
		values.background = darker(w, w->welltris.background);
	}
	if (!w->welltris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyominoColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->welltris.thingColor[thing] =
				colorCell.pixel;
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyominoColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyomino ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->welltris.reverse) {
			values.foreground = w->welltris.background;
		} else {
			values.foreground = w->welltris.foreground;
		}
	}
	if (w->welltris.thingGC[thing][2])
		XtReleaseGC((Widget) w, w->welltris.thingGC[thing][2]);
	w->welltris.thingGC[thing][2] = XtGetGC((Widget) w, valueMask, &values);
	gotColor = False;
	if (!w->welltris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyominoColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->welltris.thingColor[thing] =
				brighter(w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyominoColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyomino ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->welltris.reverse) {
			values.foreground = w->welltris.background;
		} else {
			values.foreground = w->welltris.foreground;
		}
	}
	if (w->welltris.thingGC[thing][1])
		XtReleaseGC((Widget) w, w->welltris.thingGC[thing][1]);
	w->welltris.thingGC[thing][1] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->welltris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyominoColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->welltris.thingColor[thing] =
				darker(w, colorCell.pixel);
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyominoColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyomino ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->welltris.reverse) {
			values.foreground = w->welltris.background;
		} else {
			values.foreground = w->welltris.foreground;
		}
	}
	if (w->welltris.thingGC[thing][3])
		XtReleaseGC((Widget) w, w->welltris.thingGC[thing][3]);
	w->welltris.thingGC[thing][3] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->welltris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyominoColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->welltris.thingColor[thing] =
				brighter(w, brighter(w, colorCell.pixel));
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyominoColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyomino ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->welltris.reverse) {
			values.foreground = w->welltris.background;
		} else {
			values.foreground = w->welltris.foreground;
		}
	}
	if (w->welltris.thingGC[thing][0])
		XtReleaseGC((Widget) w, w->welltris.thingGC[thing][0]);
	w->welltris.thingGC[thing][0] = XtGetGC((Widget) w, valueMask,
		&values);
	gotColor = False;
	if (!w->welltris.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				DefaultColormapOfScreen(XtScreen(w)),
				polyominoColorName[thing],
				&colorCell, &rgb)) {
			values.foreground = w->welltris.thingColor[thing] =
				darker(w, darker(w, colorCell.pixel));
			gotColor = True;
		} else {
			char *buf1, *buf2;

			stringCat(&buf1, "Color name \"",
				polyominoColorName[thing]);
			stringCat(&buf2, buf1, "\" is not defined for polyomino ");
			free(buf1);
			intCat(&buf1, buf2, thing);
			free(buf2);
			DISPLAY_WARNING(buf1);
			free(buf1);
		}
	}
	if (!gotColor) {
		if (w->welltris.reverse) {
			values.foreground = w->welltris.background;
		} else {
			values.foreground = w->welltris.foreground;
		}
	}
	if (w->welltris.thingGC[thing][4])
		XtReleaseGC((Widget) w, w->welltris.thingGC[thing][4]);
	w->welltris.thingGC[thing][4] = XtGetGC((Widget) w, valueMask,
		&values);
}

static void
setAllColors(WelltrisWidget w)
{
	XGCValues values;
	XtGCMask valueMask;
	int polyomino;

	valueMask = GCForeground | GCBackground;
	if (w->welltris.reverse) {
		values.background = w->welltris.background;
		values.foreground = w->welltris.foreground;
	} else {
		values.foreground = w->welltris.background;
		values.background = w->welltris.foreground;
	}
	if (w->welltris.inverseGC[1])
		XtReleaseGC((Widget) w, w->welltris.inverseGC[1]);
	w->welltris.inverseGC[1] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->welltris.mono) {
		values.foreground = brighter(w, (w->welltris.reverse) ?
			w->welltris.foreground : w->welltris.background);
	}
	if (w->welltris.inverseGC[0])
		XtReleaseGC((Widget) w, w->welltris.inverseGC[0]);
	w->welltris.inverseGC[0] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->welltris.mono) {
		values.foreground = darker(w, (w->welltris.reverse) ?
			w->welltris.foreground : w->welltris.background);
	}
	if (w->welltris.inverseGC[2])
		XtReleaseGC((Widget) w, w->welltris.inverseGC[2]);
	w->welltris.inverseGC[2] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->welltris.mono) {
		values.foreground = darker(w, darker(w, (w->welltris.reverse) ?
			w->welltris.foreground : w->welltris.background));
	}
	if (w->welltris.inverseGC[3])
		XtReleaseGC((Widget) w, w->welltris.inverseGC[3]);
	w->welltris.inverseGC[3] = XtGetGC((Widget) w, valueMask, &values);
	if (!w->welltris.mono) {
		values.foreground = darker(w, darker(w, darker(w,
			(w->welltris.reverse) ? w->welltris.foreground :
			w->welltris.background)));
	}
	if (w->welltris.inverseGC[4])
		XtReleaseGC((Widget) w, w->welltris.inverseGC[4]);
	w->welltris.inverseGC[4] = XtGetGC((Widget) w, valueMask, &values);
	for (polyomino = 0; polyomino < MAX_START_POLYOMINOES; polyomino++)
		getColor(w, polyomino);
	valueMask = GCForeground | GCBackground | GCFunction;
	values.foreground = 0xFFFFFF;
	values.background = 0;
	values.function = GXxor;
	if (w->welltris.xorGC)
		XtReleaseGC((Widget) w, w->welltris.xorGC);
	w->welltris.xorGC = XtGetGC((Widget) w, valueMask, &values);
}

static Boolean
setValuesGame(Widget current, Widget request, Widget renew)
{
	WelltrisWidget c = (WelltrisWidget) current, w = (WelltrisWidget) renew;
	Boolean redraw = False, setColors = False;

	if (w->welltris.act) {
		falldownGame(w);
		w->welltris.act = False;
	}
	if (w->welltris.level != c->welltris.level)
		w->welltris.delay = delay[w->welltris.level - 1];
	checkThings(w);
	if (w->welltris.reverse != c->welltris.reverse ||
			w->welltris.mono != c->welltris.mono) {
		setAllColors(w);
		redraw = True;
	} else if (w->welltris.background != c->welltris.background ||
			w->welltris.foreground != c->welltris.foreground ||
			setColors) {
		setAllColors(w);
		redraw = True;
	} else if (w->welltris.plainPiece != c->welltris.plainPiece) {
		redraw = True;
	} else if (w->welltris.suddenAppear != c->welltris.suddenAppear) {
		redoNext(w);
	}
	if (w->welltris.diameter != c->welltris.diameter) {
		w->welltris.paused = True;
		w->welltris.started = False;
		eraseFrame(w);
		sizeGame(w);
		resizeThings(w);
		newThing(w);
		newThing(w);
		clearThings(w, True);
		drawFrame(w, w->welltris.focus);
		if (w->welltris.grid)
			drawGrid(w);
		drawField(w);
		redraw = True;
	} else if (w->welltris.fill != c->welltris.fill) {
		clearThings(w, True);
		redraw = True;
	} else if (w->welltris.units != c->welltris.units) {
		w->welltris.next.units = w->welltris.units;
		w->welltris.next.bonus = False;
		w->welltris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->welltris.diagonal != c->welltris.diagonal) {
		w->welltris.next.diagonal = w->welltris.diagonal;
		w->welltris.next.bonus = False;
		w->welltris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->welltris.mixed != c->welltris.mixed) {
		w->welltris.next.mixed = w->welltris.mixed;
		w->welltris.next.bonus = False;
		w->welltris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->welltris.bonus != c->welltris.bonus) {
		w->welltris.next.bonus = False;
		w->welltris.newRows = 0;
		redoNext(w);
		drawNext(w);
	} else if (w->welltris.grid != c->welltris.grid) {
		eraseFrame(w);
		drawFrame(w, w->welltris.focus);
		if (w->welltris.grid)
			drawGrid(w);
		drawField(w);
	}
	if (w->welltris.menu != -1) {
		switch (w->welltris.menu) {
		case 1:
			w->welltris.paused = True;
			break;
		case 2:
			w->welltris.menu = -1;
			clearThings(w, True);
			redraw = True;
			break;
		case 4:
			w->welltris.menu = -1;
			/*ConfirmThings(w);*/
			redraw = True;
			break;
		default:
			w->welltris.menu = -1;
			break;
		}
	}
	return (redraw);
}

static void
quitGame(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	Display *display = XtDisplay(w);

	gameOver(w, True);
	XtCloseDisplay(display);
	exit(0);
}

static void
destroyGame(Widget old)
{
	WelltrisWidget w = (WelltrisWidget) old;
	int thing, i;

#if defined( USE_SOUND ) && defined( USE_ESOUND )
	(void) shutdown_sound();
#endif
	for (thing = 0; thing < MAX_START_POLYOMINOES; thing++) {
		for (i = 0; i < FG_SHADES; i++)
			XtReleaseGC(old, w->welltris.thingGC[thing][i]);
	}
	for (i = 0; i < BG_SHADES; i++)
		XtReleaseGC(old, w->welltris.inverseGC[i]);
	XtReleaseGC(old, w->welltris.xorGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->welltris.select);
}
#endif

#ifndef WINVER
static
#endif
void
resizeGame(WelltrisWidget w)
{
#ifdef WINVER
	RECT rect;

	/* Determine size of client area */
	(void) GetClientRect(w->core.hWnd, &rect);
	w->core.width = rect.right;
	w->core.height = rect.bottom;
#endif
	w->welltris.boxSize = MIN((w->core.height - 6),
		(w->core.width - 6)) / (3 * w->welltris.diameter);
	w->welltris.viewLength = 3 * w->welltris.boxSize *
		w->welltris.diameter;
	w->welltris.offset.x = ((int) w->core.width - 2 -
		w->welltris.viewLength) / 2;
	w->welltris.offset.y = ((int) w->core.height - 2  -
		w->welltris.viewLength) / 2;
}

#ifndef WINVER
static
#endif
void
sizeGame(WelltrisWidget w)
{
	resetThings(w);
	resizeGame(w);
}

#ifndef WINVER
static
#endif
void
initializeGame(
#ifdef WINVER
WelltrisWidget w, HBRUSH brush
#else
Widget request, Widget renew
#endif
)
{
	int diagonal, mixed, turnStyle, wall;
#ifdef IMAGE
	int bitmap;
#endif
#ifndef WINVER
	int thing, i;
	WelltrisWidget w = (WelltrisWidget) renew;
#endif

#ifdef WINVER
	setValuesGame(w);
#else
	w->welltris.mono = (DefaultDepthOfScreen(XtScreen(w)) < 2 ||
		w->welltris.mono);
	for (thing = 0; thing < MAX_START_POLYOMINOES; thing++) {
		for (i = 0; i < FG_SHADES; i++)
			w->welltris.thingGC[thing][i] = NULL;
	}
	for (i = 0; i < BG_SHADES; i++)
		w->welltris.inverseGC[i] = NULL;
	w->welltris.xorGC = NULL;
#endif
#ifdef IMAGE
	for (bitmap = 0; bitmap < BITMAPS; bitmap++)
		images[FALSE][bitmap] = None;
#endif
	w->welltris.wall = NULL;
	w->welltris.base = NULL;
	w->welltris.started = False;
	w->welltris.quit = False;
	w->welltris.completedRows = 0;
	w->welltris.newRows = 0;
	w->welltris.fall = False;
	w->welltris.focus = False;
	w->welltris.next.units = w->welltris.units;
	w->welltris.next.diagonal = w->welltris.diagonal;
	w->welltris.next.mixed = w->welltris.mixed;
	w->welltris.next.bonus = False;
	w->welltris.startLevel = w->welltris.level;
	w->welltris.nextThing.units = 0;
	w->welltris.nextThing.polyNumber = 0;
	w->welltris.nextThing.xpos = 0;
	w->welltris.nextThing.ypos = 0;
	w->welltris.nextThing.size = 0;
	w->welltris.nextThing.colorNumber = 0;
	w->welltris.nextThing.bonus = 0;
	w->welltris.nextThing.randomRotation = 0;
	w->welltris.nextThing.randomReflection = 0;
	w->welltris.nextThing.randomNumber = 0;
	w->welltris.gotNew = False;
	w->welltris.dropPaused = 0;
	checkThings(w);
	sizeGame(w);
	for (wall = 0; wall < MAX_SIDES; wall++)
		w->welltris.frozenWall[wall] = 0;
#ifdef WINVER
	brush = CreateSolidBrush(w->welltris.inverseGC[2]);
	SETBACK(w->core.hWnd, brush);
	(void) SRAND(time(NULL));
#else
	(void) SRAND(getpid());
	setAllColors(w);
#endif
	diagonal = (w->welltris.next.diagonal) ? 1 : 0;
	mixed = (w->welltris.next.mixed) ? 1 : 0;
	readThings();
	turnStyle = tris[w->welltris.next.units -
		MIN_UNITS].mode[diagonal].turnStyle;
	/* use intCat */
	if (turnStyle < 0 || turnStyle >= MAX_TURNSTYLE_TYPES) {
		(void) fprintf(stderr, "Welltris: corrupted polyomino file\n");
		(void) fprintf(stderr, "\tturnStyle = %d.\n", turnStyle);
		exit(1);
	}
	w->welltris.numberPolyominoes = tris[w->welltris.next.units -
		MIN_UNITS].mode[diagonal].number[mixed][turnStyle];
	if (w->welltris.numberPolyominoes <= 0 ||
			w->welltris.numberPolyominoes > MAX_START_POLYOMINOES) {
		(void) fprintf(stderr, "Welltris: corrupted polyomino file\n");
		(void) fprintf(stderr, "\tnumberPolyominoes = %d.\n",
			w->welltris.numberPolyominoes);
		exit(1);
	}
	clearThings(w, False);
	newThing(w);
	newThing(w);
#ifdef USE_SOUND
#ifdef USE_NAS
	dsp = XtDisplay(w);
#endif
#ifdef USE_ESOUND
	(void) init_sound();
#endif
#endif
}

#ifndef WINVER
static
#endif
void
exposeGame(
#ifdef WINVER
WelltrisWidget w
#else
Widget renew, XEvent *event, Region region
#endif
)
{
#ifndef WINVER
	WelltrisWidget w = (WelltrisWidget) renew;

	if (!w->core.visible)
		return;
#endif
	eraseFrame(w);
	resizeThings(w);
	drawFrame(w, w->welltris.focus);
	if (w->welltris.grid)
		drawGrid(w);
	drawField(w);
	drawThing(w);
	drawNext(w);
}

#ifndef WINVER
static
#endif
void
hideGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (w->welltris.sound)
		w->welltris.sound = False;
	w->welltris.paused = True;
	w->welltris.dropPaused = 2;
	setGame(w, ACTION_HIDE);
}

#ifndef WINVER
static
#endif
void
clearGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	clearThings(w, True);
}

#ifndef WINVER
static
#endif
void
enterGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	w->welltris.focus = True;
	w->welltris.paused = False;
	drawFrame(w, w->welltris.focus);
}

#ifndef WINVER
static
#endif
void
leaveGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	/* if (w->welltris.firstFall) */
	w->welltris.focus = False;
	w->welltris.paused = True;
	if (w->welltris.dropPaused == 0)
		w->welltris.dropPaused = 1;
	else
		w->welltris.dropPaused = 2;
	drawFrame(w, w->welltris.focus);
}

#ifndef WINVER
static void
mouseGame(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	if (w->welltris.dropPaused == 1) /* clicking causes enter/exit events */
		w->welltris.dropPaused = 0;
}

static void
reflectGame(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, REFLECT);
}

static void
rotateGame(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, ROTATE);
}

static void
moveGameLeft(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, MOVE_LEFT);
}

static void
moveGameRight(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, MOVE_RIGHT);
}

static void
dropGame(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	moveGameInput(w, DROP);
}

#endif

#ifdef WINVER
Boolean
#else
static void
#endif
incrementGameLevel(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->welltris.paused && w->welltris.level < MAX_LEVELS) {
		w->welltris.level++;
		w->welltris.delay = delay[w->welltris.level - 1];
		w->welltris.newRows = 0;
		drawStatus(w);
#ifdef WINVER
		return True;
#endif
	}
#ifdef WINVER
		return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
decrementGameLevel(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
#ifdef CHEAT
	if (!w->welltris.paused && w->welltris.level > MIN_LEVELS) {
		w->welltris.level--;
		w->welltris.delay = delay[w->welltris.level - 1];
		w->welltris.newRows = 0;
		drawStatus(w);
#ifdef WINVER
		return True;
#endif
	}
#endif
#ifdef WINVER
		return False;
#endif
}

#ifndef WINVER
static
#endif
void
pauseGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (w->welltris.sound)
		w->welltris.sound = False;
	w->welltris.paused = !w->welltris.paused;
	w->welltris.dropPaused = 2;
#if 0
	if (w->welltris.paused) {
		banner(MSG_PAUSED);
		(void) XIconifyWindow(display, mainWin, screen_num);
	} else {
		drawNext(w);
	}
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
incrementGameUnits(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->welltris.paused && LEGAL(w->welltris.next.units + 1,
			w->welltris.next.diagonal, w->welltris.next.mixed,
			False)) {
		setGame(w, ACTION_INC_UNITS);
#ifdef WINVER
		w->welltris.next.units = w->welltris.units;
		w->welltris.next.bonus = False;
		w->welltris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
decrementGameUnits(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->welltris.paused && LEGAL(w->welltris.next.units - 1,
			w->welltris.next.diagonal, w->welltris.next.mixed,
			False)) {
		setGame(w, ACTION_DEC_UNITS);
#ifdef WINVER
		w->welltris.next.units = w->welltris.units;
		w->welltris.next.bonus = False;
		w->welltris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
incrementGameDiameter(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->welltris.paused && w->welltris.diameter < MAX_DIAMETER) {
#ifdef WINVER
		w->welltris.paused = True;
		w->welltris.started = False;
		eraseFrame(w);
		w->welltris.diameter++;
		sizeGame(w);
		resizeThings(w);
		newThing(w);
		newThing(w);
		clearThings(w, True);
		drawFrame(w, w->welltris.focus);
		if (w->welltris.grid)
			drawGrid(w);
		return True;
#else
		setGame(w, ACTION_INC_DIAMETER);
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
decrementGameDiameter(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->welltris.paused && w->welltris.diameter > MIN_DIAMETER) {
#ifdef WINVER
		w->welltris.paused = True;
		w->welltris.started = False;
		eraseFrame(w);
		w->welltris.diameter--;
		sizeGame(w);
		resizeThings(w);
		newThing(w);
		newThing(w);
		clearThings(w, True);
		drawFrame(w, w->welltris.focus);
		if (w->welltris.grid)
			drawGrid(w);
		return True;
#else
		setGame(w, ACTION_DEC_DIAMETER);
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifndef WINVER
static
#endif
void
fixedfillGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_FIXED_FILL);
}

#ifndef WINVER
static
#endif
void
suddenAppearGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_SUDDEN_APPEAR);
#ifdef WINVER
	redoNext(w);
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
diagonalGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->welltris.paused && LEGAL(w->welltris.next.units,
			!w->welltris.next.diagonal, w->welltris.next.mixed,
			False)) {
		setGame(w, ACTION_DIAGONAL);
#ifdef WINVER
		w->welltris.next.diagonal = w->welltris.diagonal;
		w->welltris.next.bonus = False;
		w->welltris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifdef WINVER
Boolean
#else
static void
#endif
mixedGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->welltris.paused && LEGAL(w->welltris.next.units,
			w->welltris.next.diagonal, !w->welltris.next.mixed,
			False)) {
		setGame(w, ACTION_MIXED);
#ifdef WINVER
		w->welltris.next.mixed = w->welltris.mixed;
		w->welltris.next.bonus = False;
		w->welltris.newRows = 0;
		redoNext(w);
		drawNext(w);
		return True;
#endif
	}
#ifdef WINVER
	return False;
#endif
}

#ifndef WINVER
static
#endif
void
bonusGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_BONUS);
#ifdef WINVER
	w->welltris.next.bonus = False;
	w->welltris.newRows = 0;
	redoNext(w);
	drawNext(w);
#endif
}

#ifndef WINVER
static
#endif
void
cwGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_CW);
}

#ifndef WINVER
static
#endif
void
plainPieceGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_PLAIN_PIECE);
#ifdef WINVER
	resizeThings(w);
	drawField(w);
#endif
}

#ifndef WINVER
static
#endif
void
showNextGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->welltris.paused) {
		setGame(w, ACTION_SHOW_NEXT);
	}
}

#ifndef WINVER
static
#endif
void
gridGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	if (!w->welltris.paused) {
		setGame(w, ACTION_GRID);
#ifdef WINVER
		eraseFrame(w);
		drawFrame(w, w->welltris.focus);
		if (w->welltris.grid)
			drawGrid(w);
		drawField(w);
#endif
	}
}

#ifndef WINVER
static
#endif
void
soundGame(WelltrisWidget w
#ifndef WINVER
, XEvent *event, char **args, int nArgs
#endif
)
{
	setGame(w, ACTION_SOUND);
}

#ifndef WINVER
static
#endif
void
fillGame(WelltrisWidget w, int n)
{
	setGame(w, n + ACTION_FILL0);
#ifdef WINVER
	clearThings(w, True);
#endif
}

#ifndef WINVER
static void
fill0Game(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 0);
}

static void
fill1Game(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 1);
}

static void
fill2Game(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 2);
}

static void
fill3Game(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 3);
}

static void
fill4Game(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 4);
}

static void
fill5Game(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 5);
}

static void
fill6Game(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 6);
}

static void
fill7Game(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 7);
}

static void
fill8Game(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 8);
}

static void
fill9Game(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 9);
}

static void
fill10Game(WelltrisWidget w, XEvent *event, char **args, int nArgs)
{
	fillGame(w, 10);
}
#endif
